__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time
import numpy as np
import pandas as pd
from tresor.util.random.Number import Number as rannum
from tresor.util.sequence.symbol.Single import Single as dnasgl
from tresor.util.Console import Console


class Error:

    def __init__(
            self,
            method,
    ):
        self.method = method
        self.console = Console()

    def __call__(self, deal):
        from functools import wraps
        if self.method == 'default':
            func = self.postable
        @wraps(deal)
        def indexing(ph, *args, **kwargs):
            res2p = deal(ph, **kwargs)
            # print(func(res2p))
            # print(kwargs['params'].keys())

            if 'verbose' in kwargs['params'].keys():
                self.console.verbose = kwargs['params']['verbose']
            else:
                self.console.verbose = True
            self.console.print('======>Generate sequencing errors...')

            return func(res2p)
        return indexing

    def postable(self, res2p, kind='index_by_lambda'):
        # print(res2p.keys())
        seq_stime = time.time()
        data_seq = pd.DataFrame(res2p['data_spl'], columns=['read', 'sam_id', 'source'])
        del res2p['data']
        del res2p['data_spl']
        self.console.print('=========>There are {} reads to be sequenced'.format(data_seq.shape[0]))
        self.console.print('=========>The position table construction starts')
        pcr_postable_stime = time.time()
        # print(data_seq['read'][0])
        if kind == 'index_by_same_len':
            seq_pos_ids, seq_ids = self.postableIndexBySameLen(
                seq_len=len(data_seq['read'][0]),
                num_seq=data_seq.shape[0],
            )
        elif kind == 'index_by_lambda':
            seq_ids = []
            seq_pos_ids = []
            data_seq.apply(lambda x: self.postableLambda(x, seq_ids, seq_pos_ids), axis=1)
        else:
            seq_pos_ids, seq_ids = self.postableIndexBySameLen(
                seq_len=len(data_seq['read'][0]),
                num_seq=data_seq.shape[0],
            )
        pos_table = {'seq_ids': seq_ids, 'seq_pos_ids': seq_pos_ids}
        # print(pos_table)
        # print(len(seq_ids))
        pcr_postable_etime = time.time()
        self.console.print('=========>Time for constructing the position table: {time:.3f}s'.format(time=pcr_postable_etime - pcr_postable_stime))
        seq_nt_num = len(seq_ids)
        self.console.print('=========>There are {} nucleotides to be sequenced'.format(seq_nt_num))
        self.console.print('=========>Determination of number of sequencing errors start')
        seq_err_num_simu_stime = time.time()
        if res2p['err_num_met'] == 'binomial':
            seq_err_num = rannum().binomial(n=seq_nt_num, p=res2p['seq_error'], use_seed=True, seed=1)
        elif res2p['err_num_met'] == 'nbinomial':
            seq_err_num = rannum().nbinomial(
                n=seq_nt_num * (1 - res2p['seq_error']),
                p=1 - res2p['seq_error'],
                use_seed=True,
                seed=1
            )
        else:
            seq_err_num = rannum().binomial(n=seq_nt_num, p=res2p['seq_error'], use_seed=True, seed=1)
        self.console.print('============>There are {} nucleotide errors during sequencing'.format(seq_err_num))
        err_lin_ids = rannum().uniform(low=0, high=seq_nt_num, num=seq_err_num, use_seed=True, seed=1)
        # print(err_lin_ids)
        arr_err_pos = []# [[row1, col1], [row2, col2], ...]
        for i in err_lin_ids:
            arr_err_pos.append([pos_table['seq_ids'][i], pos_table['seq_pos_ids'][i]])
        pseudo_nums = rannum().uniform(low=0, high=3, num=seq_err_num, use_seed=False)
        # print(pseudo_nums)
        seq_err_num_simu_etime = time.time()
        self.console.print('=========>Time for determining sequencing errors: {time:.3f}s'.format(time=seq_err_num_simu_etime - seq_err_num_simu_stime))
        self.console.print('=========>Sequencing error Assignment starts')
        seq_err_assign_stime = time.time()
        data_seq['read'] = data_seq.apply(lambda x: list(x['read']), axis=1)
        for pos_err, pseudo_num in zip(arr_err_pos, pseudo_nums):
            # print(pos_err[0])
            # print(pos_err[1])
            pcr_err_base = data_seq.loc[pos_err[0], 'read'][pos_err[1]]
            dna_map = dnasgl().todict(
                nucleotides=dnasgl().getEleTrimmed(
                    ele_loo=pcr_err_base,
                    universal=True,
                ),
                reverse=True,
            )
            # print('before', data_seq.loc[pos_err[0], 'read'][pos_err[1]])
            data_seq.loc[pos_err[0], 'read'][pos_err[1]] = dna_map[pseudo_num]
            # print('after', data_seq.loc[pos_err[0], 'read'][pos_err[1]])
        del arr_err_pos
        del pseudo_nums
        seq_err_assign_etime = time.time()
        self.console.print('=========>Time for assigning sequencing errors: {time:.2f}s'.format(time=seq_err_assign_etime - seq_err_assign_stime))
        data_seq['read'] = data_seq.apply(lambda x: ''.join(x['read']), axis=1)
        if res2p['seq_deletion']:
            data_seq['read'] = data_seq['read'].apply(lambda x: self.deletion(read=x, del_rate=res2p['seq_del_rate']))
        if res2p['seq_insertion']:
            data_seq['read'] = data_seq['read'].apply(lambda x: self.insertion(read=x, ins_rate=res2p['seq_ins_rate']))
        res2p['data'] = data_seq.values
        del data_seq
        seq_etime = time.time()
        self.console.print('=========>Sequencing time: {time:.2f}s'.format(time=seq_etime - seq_stime))
        return res2p

    def deletion(self, read, del_rate):
        num_err_per_read = rannum().binomial(
            n=len(read), p=del_rate, use_seed=False, seed=False
        )
        pos_list = rannum().choice(
            high=len(read), num=num_err_per_read, use_seed=False, seed=False, replace=False,
        )
        for _, pos in enumerate(pos_list):
            read = read[:pos] + read[pos + 1:]
        return read

    def insertion(self, read, ins_rate):
        num_err_per_read = rannum().binomial(
            n=len(read), p=ins_rate, use_seed=False, seed=False
        )
        pos_list = rannum().choice(
            high=len(read), num=num_err_per_read, use_seed=False, seed=False, replace=False,
        )
        base_list = rannum().uniform(
            low=0, high=4, num=num_err_per_read, use_seed=False
        )
        for i, pos in enumerate(pos_list):
            dna_map = dnasgl().todict(
                nucleotides=dnasgl().get(
                    universal=True,
                ),
                reverse=True,
            )
            read = read[:pos] + dna_map[base_list[i]] + read[pos:]
            ### read
            ### pos, base_list[i], dna_map[base_list[i]], dna_map
            ### read
            # 5 3 G {0: 'A', 1: 'T', 2: 'C', 3: 'G'}
            # TTTTTTTTTGGGCCCGGGAAAAAACCCAAAGGGGGG
            # TTTTTGTTTTGGGCCCGGGAAAAAACCCAAAGGGGGG
            # 9 0 A {0: 'A', 1: 'T', 2: 'C', 3: 'G'}
            # CCCTTTCCCTTTGGGTTTGGGTTTCCCGGGAAACCC
            # CCCTTTCCCATTTGGGTTTGGGTTTCCCGGGAAACCC
            # 3 0 A {0: 'A', 1: 'T', 2: 'C', 3: 'G'}
            # AAATTTTTTAAACCCAAAAAAAAAAAATTTTTTCCC
            # AAAATTTTTTAAACCCAAAAAAAAAAAATTTTTTCCC
        return read

    def postableIndexBySameLen(self, seq_len, num_seq):
        nt_ids = [i for i in range(seq_len)]
        seq_pos_ids = nt_ids * num_seq
        seq_ids = np.array([[i] * seq_len for i in range(num_seq)]).ravel().tolist()
        return seq_pos_ids, seq_ids

    def postableLambda(self, x, ids, pos_ids):
        l = len(x['read'])
        for i in range(l):
            pos_ids.append(i)
        for i in [x.name] * l:
            ids.append(i)
        return ids, pos_ids

    def todict(self, bases, reverse=False):
        aa_dict = {}
        for k, v in enumerate(bases):
            aa_dict[v] = k
        if reverse:
            aa_dict = {v: k for k, v in aa_dict.items()}
        return aa_dict


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import os


def root_dict():
    """
    ..  @description:
        -------------
        abs file path.

    :return:
    """
    ROOT_DICT = os.path.dirname(os.path.abspath(__file__))
    return ROOT_DICT


def to(path):
    """

    Parameters
    ----------
    path

    Returns
    -------

    """
    return os.path.join(
        root_dict(),
        path
    )


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time
import numpy as np
import pandas as pd
from tresor.util.random.Sampling import Sampling as ranspl
from tresor.util.random.Number import Number as rannum
from tresor.util.file.Writer import Writer as pfwriter
from tresor.util.file.Folder import Folder as crtfolder
from tresor.util.sequence.symbol.Single import Single as dnasgl
from tresor.util.Hamming import Hamming
from tresor.read.umi.Design import Design as dumi
from tresor.read.seq.Design import Design as dseq
from tresor.read.primer.Design import Design as dprimer
from tresor.read.adapter.Design import Design as dadapter
from tresor.read.spacer.Design import Design as dspacer
from tresor.util.sequence.Fasta import Fasta as sfasta
from tresor.read.error.Library import Library as errlib
from tresor.util.Kit import tactic6
from tresor.util.Console import Console


class SingleLocus:

    def __init__(
            self,
            len_params,
            seq_num=50,
            is_seed=False,

            working_dir='./simu/',
            condis=['umi'],
            sim_thres=3,
            permutation=0,
            is_sv_umi_lib=True,
            is_sv_seq_lib=True,
            is_sv_primer_lib=True,
            is_sv_adapter_lib=True,
            is_sv_spacer_lib=True,
            verbose=True,
            **kwargs,
    ):
        self.pfwriter = pfwriter()
        self.ranspl = ranspl()
        self.rannum = rannum()
        self.dnasgl = dnasgl()
        self.errlib = errlib()
        self.crtfolder = crtfolder()
        self.dumi = dumi
        self.dseq = dseq
        self.dprimer = dprimer
        self.dadapter = dadapter
        self.dspacer = dspacer
        self.working_dir = working_dir
        self.len_params = len_params
        self.is_seed = is_seed
        self.is_sv_umi_lib = is_sv_umi_lib
        self.is_sv_seq_lib = is_sv_seq_lib
        self.is_sv_primer_lib = is_sv_primer_lib
        self.is_sv_adapter_lib = is_sv_adapter_lib
        self.is_sv_spacer_lib = is_sv_spacer_lib
        self.seq_num = seq_num
        self.condis = condis
        self.sim_thres = sim_thres
        self.permutation = permutation
        self.dna_map = self.dnasgl.todict(nucleotides=self.dnasgl.get(universal=True), reverse=True)
        self.crtfolder.osmkdir(working_dir)

        self.kwargs = kwargs

        if not self.kwargs['material_params']:
            self.kwargs['material_params'] = {}
            self.kwargs['material_params']['fasta_cdna_fpn'] = None

        self.console = Console()
        self.console.verbose = verbose

        self.console.print('Initialisation and parameters: \n{}'.format(self.kwargs))

    def pooling(self, ):
        """

        Attributes
        ----------
        condi_map
            {'umi': ['alone', '1'], 'primer': ['alone', '1'], 'spacer': ['alone', '1'], 'adapter': ['alone', '1'], 'seq': ['alone', '2']}

        Returns
        -------

        """
        self.console.print("======>Sequencing library preparation starts")
        stime = time.time()
        sequencing_library = []
        mut_recorder_arr = []
        del_recorder_arr = []
        ins_recorder_arr = []
        umi_pool = []
        umi_cnt = 0

        ### +++++++++++++++ block: condition map +++++++++++++++
        condi_map = {}
        for condi in self.condis:
            condi_arr = condi.split("_")
            condi_map[condi_arr[0]] = []
        for condi in self.condis:
            condi_arr = condi.split("_")
            if len(condi_arr) == 1:
                condi_map[condi_arr[0]].append('alone')
            else:
                condi_map[condi_arr[0]].append(condi_arr[1])
        self.console.print("======>Condition map: {}".format(condi_map))
        condi_keys = condi_map.keys()

        ### +++++++++++++++ block: select CDNA from a reference ome +++++++++++++++
        if 'seq' in condi_keys and self.kwargs['material_params']['fasta_cdna_fpn']:
            self.console.print("======>Read CDNAs from a reference ome")
            fastq_ref_arr = sfasta().get_from_gz(
                fasta_fpn=self.kwargs['material_params']['fasta_cdna_fpn'],
            )
            if self.kwargs['mode'] == 'short_read':
                df_fastq_ref = pd.DataFrame(fastq_ref_arr)
                # print(df_fastq_ref)
                df_fastq_ref = df_fastq_ref.rename(columns={0: 'name', 1: 'seq'})
                df_fastq_ref['len'] = df_fastq_ref['seq'].apply(lambda x: len(x))
                df_fastq_ref = df_fastq_ref.loc[df_fastq_ref['len'] > 2 * self.len_params['seq']]
                # print(df_fastq_ref)
                seq_cdna_map = tactic6(
                    arr_2d=df_fastq_ref[['name', 'seq']].values.tolist()
                )
                cdna_ids = [*seq_cdna_map.keys()]
            else:
                seq_cdna_map = tactic6(
                    arr_2d=fastq_ref_arr
                )
                cdna_ids = [*seq_cdna_map.keys()]
            self.console.print("======>There are {} genes in the reference genome".format(len(cdna_ids)))

            cdna_seqs_sel_maps = {}
            for i, seq_i in enumerate(condi_map['seq']):
                cdna_ids_sel = self.ranspl.uniform(
                    data=cdna_ids,
                    num=self.seq_num,
                    use_seed=self.is_seed,
                    seed=i + 100000,
                    replace=True,
                )
                ### cdna_ids_sel
                # ['ENST00000526369.3', ..., 'ENST00000531461.5', 'ENST00000635785.1']
                # ['ENST00000581517.1', ..., 'ENST00000490527.1', 'ENST00000453971.1']
                # ...
                # ['ENST00000519955.1', ..., 'ENST00000527746.5', 'ENST00000360135.8']
                u = []
                for ii in cdna_ids_sel:
                    if self.kwargs['mode'] == 'long_read':
                        u.append(seq_cdna_map[ii])
                    elif self.kwargs['mode'] == 'short_read':
                        # @@ if a random number extending to the end of the reference CDNA sequence
                        # leads to a length of sequence, which is shorter than the required short
                        # read length, then we directly use the whole reference CDNA sequence.
                        cdna_short_read_ran_id = self.rannum.uniform(
                            low=0,
                            high=len(seq_cdna_map[ii]) - self.len_params['seq'],
                            num=1,
                            use_seed=self.is_seed,
                            seed=i + 1,
                        )
                        if (cdna_short_read_ran_id[0] + self.len_params['seq']) > len(seq_cdna_map[ii]):
                            u.append(seq_cdna_map[ii][0:self.len_params['seq']])
                        else:
                            u.append(seq_cdna_map[ii][cdna_short_read_ran_id[0]:(cdna_short_read_ran_id[0] + self.len_params['seq'])])
                cdna_seqs_sel_maps[seq_i] = u
                # print(cdna_seqs_sel_maps)
                # print(len(cdna_seqs_sel_maps))
                self.pfwriter.generic(df=cdna_ids_sel, sv_fpn=self.working_dir + 'cdna_ids_' + seq_i + '.txt')
            del seq_cdna_map

        ### +++++++++++++++ block: generate each read +++++++++++++++
        for id in np.arange(self.seq_num):
            self.console.print("======>Read {} generation".format(id + 1))
            read_struct_ref = {}
            ### +++++++++++++++ block: generate umis +++++++++++++++
            if 'umi' in condi_keys:
                self.console.print("=========>UMI generation start")
                for umi_mark_id, umi_mark_suffix in enumerate(condi_map['umi']):
                    # print(umi_mark_id, id + self.permutation * self.seq_num + umi_cnt + umi_mark_id + 100000000)
                    umi_mark = '_' + umi_mark_suffix if umi_mark_suffix != 'alone' else ''
                    self.console.print("============>UMI condition {}: {}".format(umi_mark_id, 'umi' + umi_mark))
                    umi_flag = False
                    while not umi_flag:
                        umi_seed = id + self.permutation * self.seq_num + umi_cnt + (umi_mark_id + 1) * 100000000
                        umip = self.dumi(
                            dna_map=self.dna_map,
                            umi_unit_pattern=self.len_params['umi' + umi_mark]['umi_unit_pattern'],
                            pseudorandom_num=self.rannum.uniform(
                                low=0,
                                high=4,
                                num=self.len_params['umi' + umi_mark]['umi_unit_len'],
                                use_seed=self.is_seed,
                                seed=umi_seed,
                            ),
                        )
                        umi_i = umip.reoccur(is_sv=False)
                        edh = np.array([Hamming().general(umi_i, j) for j in umi_pool])
                        # for j in umi_pool:
                        #     if hamming().general(umi_i, j) < self.sim_thres:
                        #         print(umi_i, j)
                        if len(edh[edh < self.sim_thres]) == 0:
                            # print(len(edh[edh < self.sim_thres]))
                            umi_pool.append(umi_i)
                            read_struct_ref['umi' + umi_mark] = umi_i
                            umi_flag = True
                            umip.write(res=umi_i, lib_fpn=self.working_dir + 'umi' + umi_mark + '.txt', is_sv=self.is_sv_umi_lib)
                            umip.write(
                                res=str(umi_seed),
                                lib_fpn=self.working_dir + 'umi' + umi_mark + '_seeds.txt',
                                is_sv=self.is_sv_umi_lib,
                            )
                        else:
                            # print(id)
                            umi_cnt += 1

            ### +++++++++++++++ block: generate seqs +++++++++++++++
            if 'seq' in condi_keys:
                self.console.print("=========>Sequence generation start")
                for seq_mark_id, seq_mark_suffix in enumerate(condi_map['seq']):
                    seq_mark = '_' + seq_mark_suffix if seq_mark_suffix != 'alone' else ''
                    self.console.print("============>Sequence condition {}: {}".format(seq_mark_id, 'seq' + seq_mark))
                    if self.kwargs['material_params']['fasta_cdna_fpn']:
                        seq_i = self.dseq(
                            cdna_seq=cdna_seqs_sel_maps[seq_mark_suffix][id],
                        ).cdna(lib_fpn=self.working_dir + 'seq' + seq_mark + '.txt', is_sv=self.is_sv_seq_lib)
                    else:
                        seq_seed = id + self.permutation * self.seq_num + 8000000 + (seq_mark_id+1) * 200000000
                        pseq = self.dseq(
                            dna_map=self.dna_map,
                            pseudorandom_num=self.rannum.uniform(
                                low=0,
                                high=4,
                                num=self.len_params['seq' + seq_mark],
                                use_seed=self.is_seed,
                                seed=seq_seed,
                            ),
                        )
                        seq_i = pseq.general(lib_fpn=self.working_dir + 'seq' + seq_mark + '.txt', is_sv=self.is_sv_seq_lib)
                        pseq.write(
                            res=str(seq_seed),
                            lib_fpn=self.working_dir + 'seq' + seq_mark + '_seeds.txt',
                            is_sv=self.is_sv_seq_lib,
                        )
                    read_struct_ref['seq' + seq_mark] = seq_i

            ### +++++++++++++++ block: generate primers +++++++++++++++
            if 'primer' in condi_keys:
                self.console.print("=========>Primer generation start")
                for primer_mark_id, primer_mark_suffix in enumerate(condi_map['primer']):
                    primer_mark = '_' + primer_mark_suffix if primer_mark_suffix != 'alone' else ''
                    self.console.print("============>Primer condition {}: {}".format(primer_mark_id, 'primer' + primer_mark))
                    primer_seed = id + self.permutation * self.seq_num + 8000000 + (primer_mark_id + 1) * 300000000
                    pprimer = self.dprimer(
                        dna_map=self.dna_map,
                        pseudorandom_num=self.rannum.uniform(
                            low=0,
                            high=4,
                            num=self.len_params['primer' + primer_mark],
                            use_seed=self.is_seed,
                            seed=primer_seed,
                        ),
                    )
                    primer_i = pprimer.general(lib_fpn=self.working_dir + 'primer' + primer_mark + '.txt', is_sv=self.is_sv_primer_lib)
                    pprimer.write(
                        res=str(primer_seed),
                        lib_fpn=self.working_dir + 'primer' + primer_mark + '_seeds.txt',
                        is_sv=self.is_sv_primer_lib,
                    )
                    read_struct_ref['primer' + primer_mark] = primer_i

            ### +++++++++++++++ block: generate adapters +++++++++++++++
            if 'adapter' in condi_keys:
                self.console.print("=========>Adapter generation start")
                for adapter_mark_id, adapter_mark_suffix in enumerate(condi_map['adapter']):
                    adapter_mark = '_' + adapter_mark_suffix if adapter_mark_suffix != 'alone' else ''
                    self.console.print("============>Adapter condition {}: {}".format(adapter_mark_id, 'adapter' + adapter_mark))
                    adapter_seed = id + self.permutation * self.seq_num + 8000000 + (adapter_mark_id+1) * 400000000
                    padapter = self.dadapter(
                        dna_map=self.dna_map,
                        pseudorandom_num=self.rannum.uniform(
                            low=0,
                            high=4,
                            num=self.len_params['adapter' + adapter_mark],
                            use_seed=self.is_seed,
                            seed=adapter_seed,
                        ),
                    )
                    adapter_i = padapter.general(lib_fpn=self.working_dir + 'adapter' + adapter_mark + '.txt', is_sv=self.is_sv_adapter_lib)
                    padapter.write(
                        res=str(adapter_seed),
                        lib_fpn=self.working_dir + 'adapter' + adapter_mark + '_seeds.txt',
                        is_sv=self.is_sv_adapter_lib,
                    )
                    read_struct_ref['adapter' + adapter_mark] = adapter_i

            ### +++++++++++++++ block: generate spacers +++++++++++++++
            if 'spacer' in condi_keys:
                self.console.print("=========>Spacer generation start")
                for spacer_mark_id, spacer_mark_suffix in enumerate(condi_map['spacer']):
                    spacer_mark = '_' + spacer_mark_suffix if spacer_mark_suffix != 'alone' else ''
                    self.console.print("============>Spacer condition {}: {}".format(spacer_mark_id, 'spacer' + spacer_mark))
                    spacer_seed = id + self.permutation * self.seq_num + 8000000 + (spacer_mark_id+1) * 500000000
                    pspacer = self.dspacer(
                        dna_map=self.dna_map,
                        pseudorandom_num=self.rannum.uniform(
                            low=0,
                            high=4,
                            num=self.len_params['spacer' + spacer_mark],
                            use_seed=self.is_seed,
                            seed=spacer_seed,
                        ),
                    )
                    spacer_i = pspacer.general(lib_fpn=self.working_dir + 'spacer' + spacer_mark + '.txt', is_sv=self.is_sv_spacer_lib)
                    pspacer.write(
                        res=str(spacer_seed),
                        lib_fpn=self.working_dir + 'spacer' + spacer_mark + '_seeds.txt',
                        is_sv=self.is_sv_spacer_lib,
                    )
                    read_struct_ref['spacer' + spacer_mark] = spacer_i

            ### +++++++++++++++ block: Custom-designed sequences +++++++++++++++
            if 'custom' in condi_keys:
                for custom_mark_id, custom_mark_suffix in enumerate(condi_map['custom']):
                    custom_mark = '_' + custom_mark_suffix if custom_mark_suffix != 'alone' else ''
                    self.console.print("============>Custom-designed condition {}: {}".format(custom_mark_id, 'custom' + custom_mark))
                    read_struct_ref['custom' + custom_mark] = self.kwargs['seq_params']['custom' + custom_mark]
            # print(read_struct_ref)
            read_struct_pfd_order = {condi: read_struct_ref[condi] for condi in self.condis}
            read, read_err_dict = self.paste([*read_struct_pfd_order.values()])
            sequencing_library.append([read, str(id), 'init'])
            # print(read_err_dict['read_mut']['mark'])
            # print(read_err_dict)
            # print(read_err_dict)
            if 'read_mut' in read_err_dict.keys():
                mut_recorder_arr.append(read_err_dict['read_mut']['mark'])
            if 'read_del' in read_err_dict.keys():
                del_recorder_arr.append(read_err_dict['read_del']['mark'])
            if 'read_ins' in read_err_dict.keys():
                ins_recorder_arr.append(read_err_dict['read_ins']['mark'])
        # print(umi_cnt)
        # print(umi_pool)
        # print(sequencing_library)
        # print(mut_recorder_arr)
        # print(del_recorder_arr)
        # print(ins_recorder_arr)
        df = pd.DataFrame({
            'mut': mut_recorder_arr,
            'del': del_recorder_arr,
            'ins': ins_recorder_arr,
        })
        # print(df)
        # print(sum(mut_recorder_arr))
        # print(sum(del_recorder_arr))
        # print(sum(ins_recorder_arr))
        self.pfwriter.generic(df=sequencing_library, sv_fpn=self.working_dir + 'sequencing_library.txt')
        etime = time.time()
        self.console.print("===>Time for sequencing library preparation: {:.3f}s".format(etime-stime))
        return sequencing_library, df

    def paste(self, read_struct=[]):
        read = ''.join(read_struct)
        if 'bead_mutation' in self.kwargs.keys() and self.kwargs['bead_mutation']:
            # print('bead_mutation')
            # print(read)
            read, read_mut = self.errlib.mutated(
                read=read,
                pcr_error=self.kwargs['bead_mut_rate'],
                mode='bead_mutation',
            )
            # print(read)
            # print(read_mut)
        else:
            read_mut = {}
        if 'bead_deletion' in self.kwargs.keys() and self.kwargs['bead_deletion']:
            # print('bead_deletion')
            # print(read)
            read, read_del = self.errlib.deletion(
                read=read,
                del_rate=self.kwargs['bead_del_rate'],
                mode='bead_deletion',
            )
            # print(read)
        else:
            read_del = {}
        if 'bead_insertion' in self.kwargs.keys() and self.kwargs['bead_insertion']:
            # print('bead_insertion')
            # print(read)
            read, read_ins = self.errlib.insertion(
                read=read,
                ins_rate=self.kwargs['bead_ins_rate'],
                mode='bead_insertion',
            )
            # print(read)
        else:
            read_ins = {}
        return read, {
            'read_mut': read_mut,
            'read_del': read_del,
            'read_ins': read_ins,
        }


if __name__ == "__main__":
    from tresor.path import to
    # print(DEFINE['cand_pool_fpn'])
    p = SingleLocus(
        seq_num=50,
        len_params={
            'umi': {
                'umi_unit_pattern': 1,
                'umi_unit_len': 12,
            },
            'umi_1': {
                'umi_unit_pattern': 3,
                'umi_unit_len': 12,
            },
            'seq': 100,
            'seq_2': 100,
            'adapter': 10,
            'adapter_1': 10,
            'primer': 10,
            'primer_1': 10,
            'spacer': 10,
            'spacer_1': 10,
        },
        seq_params={
            'custom': 'AAGC', # BAGC
            'custom_1': 'V',
        },
        material_params={
            # 'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
        },
        is_seed=True,

        working_dir=to('data/simu/'),

        # condis=['umi'],
        condis=['umi', 'seq'],
        # condis=['umi', 'custom', 'seq', 'custom_1'],
        # condis=['umi', 'primer', 'primer_1', 'spacer', 'spacer_1', 'adapter', 'adapter_1', 'seq', 'seq_2', 'umi_1'],
        sim_thres=3,
        permutation=0,

        bead_mutation=True,  # True False
        bead_mut_rate=1e-4,  # 0.016 0.00004
        bead_deletion=True,  # True False
        bead_insertion=True,
        bead_del_rate=0.1/112,  # 0.016 0.00004, 2.4e-7
        bead_ins_rate=7.1e-7,  # 0.011 0.00001, 7.1e-7

        mode='short_read',  # long_read short_read bead_synthesis

        verbose=True, # False True
    )

    # print(p.umi_len)
    p.pooling()


from importlib import metadata as importlib_metadata

from . import path
from . import gmat
from . import gene
from . import gcell
from . import locus
from . import sc

# __all__ = [
#     "gene",
#     "gcell",
#     "locus",
#     "sc",
# ]


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import numpy as np
from tresor.pcr.Amplify import Amplify as pcr
from tresor.sequencing.Calling import Calling as seq
from tresor.read.error.Library import Library as errlib
from tresor.util.sequence.fastq.Write import write as wfastq
from tresor.util.file.Reader import Reader as pfreader
from tresor.util.random.Number import Number as rannum
from tresor.util.random.Sampling import Sampling as ranspl
from tresor.util.Console import Console
from tresor.util.Kit import tactic6


class Subsampling:

    def __init__(
            self,
            verbose=False
    ):
        self.pcr = pcr
        self.seq = seq
        self.wfastq = wfastq
        self.rannum = rannum
        self.ranspl = ranspl
        self.errlib = errlib()
        self.console = Console()
        self.console.verbose = verbose

    def mutation_table_minimum(
            self,
            pcr_dict,
            replace=False,
    ):
        """

        Parameters
        ----------
        pcr_dict

        Returns
        -------

        """
        self.console.print('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
        self.console.print('======>Substitutions of nucleotides by PCR errors using mutation_table_minimum')
        self.console.print('======>Read PCR amplified reads')
        df_seq_lib = pfreader().generic(pcr_dict['read_lib_fpn'])
        # umi_map = df_seq_lib[0].to_dict()
        import pandas as pd
        umi_map = pd.Series(df_seq_lib[0].values, index=df_seq_lib[1].astype(str)).to_dict()
        # print(umi_map)

        self.console.print('======>Sampling reads to be sequenced')
        num_all_pcr_ampl_reads = pcr_dict['mut_info'].shape[0]
        self.console.print(
            '=========>There are a total number of {} PCR amplified reads'.format(num_all_pcr_ampl_reads))
        # print(num_all_pcr_ampl_reads)
        # print(pcr_dict.keys())
        if pcr_dict['seq_sub_spl_number'] is not None:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_number']
        else:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_rate'] * num_all_pcr_ampl_reads
        # spl_ids = rannum().choice(
        #     # low=0,
        #     high=num_all_pcr_ampl_reads,
        #     num=num_reads_for_sequencing,
        #     use_seed=pcr_dict['use_seed'],
        #     seed=pcr_dict['seed'],
        #     replace=replace,
        # )
        spl_ids = rannum().uniform(
            low=0, high=num_all_pcr_ampl_reads, num=num_reads_for_sequencing, use_seed=False, seed=1
        )
        # print(spl_ids)
        # print(pcr_dict['data'])
        spl_id_map = tactic6(pcr_dict['data'][:, [1, 2]])
        # print(pcr_dict['data'])
        # print(spl_id_map)
        # print(len(spl_id_map))
        spl_mut_info = pcr_dict['mut_info'][spl_ids]
        # print(pcr_dict['mut_info'].shape)
        # print(spl_mut_info)
        # print(len(spl_mut_info))
        keys = spl_mut_info[:, 2]
        # keys = pcr_dict['data'][spl_ids][:, 1]
        # print(keys)
        # print(np.sort(keys) == np.sort(spl_mut_info[:, 2]))
        # print(len(keys))
        pos_dict = tactic6(pcr_dict['mut_info'][:, [2, 0]])
        base_dict = tactic6(pcr_dict['mut_info'][:, [2, 1]])
        # print(pos_dict)
        # print(base_dict)
        # print(tactic6(base_np))
        res_data = []
        for key in keys:
            mol_id = key.split('_')[0]
            k = key.split('_')[1:]
            # print('kkk', key, k)
            read = umi_map[mol_id]
            # print(read)
            for i in range(len(k)):
                # print('id', i)
                sub_k = mol_id + '_' + '_'.join(k[: i + 1]) if k != [] else mol_id
                # print(sub_k)
                # print(pos_dict[sub_k], base_dict[sub_k])
                if sub_k in pos_dict.keys():
                    read = self.errlib.change(read, pos_list=pos_dict[sub_k], base_list=base_dict[sub_k])
            #     print(read)
            # print(read)
            res_data.append([
                read,  # read
                str(mol_id) + '_' + '_'.join(k) if k != [] else str(mol_id),  # sam id
                spl_id_map[str(mol_id) + '_' + '_'.join(k)] if k != [] else 'init',  # source
            ])
            # print(read)
        # print(np.array(res_data).shape)
        return np.array(res_data)

    def mutation_table_complete(
            self,
            pcr_dict,
            replace=False,
    ):
        # print(pcr_dict)
        # print(pcr_dict['mut_info'])
        self.console.print('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
        self.console.print('======>Substitutions of nucleotides by PCR errors using mutation_table_complete')
        self.console.print('======>Read PCR amplified reads')
        df_seq_lib = pfreader().generic(pcr_dict['read_lib_fpn'])
        # umi_map = df_seq_lib[0].to_dict()
        import pandas as pd
        umi_map = pd.Series(df_seq_lib[0].values, index=df_seq_lib[1].astype(str)).to_dict()
        # print(umi_map)
        # num_all_pcr_ampl_reads = pcr_dict['data'].shape[0]
        # self.console.print('=========>There are a total number of {} PCR amplified reads'.format(num_all_pcr_ampl_reads))
        # spl_ids = rannum().uniform(
        #     low=0, high=num_all_pcr_ampl_reads, num=9000, use_seed=False, seed=1
        # )
        self.console.print('======>Sampling reads to be sequenced')
        num_all_pcr_ampl_reads = pcr_dict['data'].shape[0]
        self.console.print('=========>There are a total number of {} PCR amplified reads'.format(num_all_pcr_ampl_reads))
        # print(num_all_pcr_ampl_reads)
        # print(pcr_dict.keys())
        if pcr_dict['seq_sub_spl_number'] is not None:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_number']
        else:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_rate'] * num_all_pcr_ampl_reads
        spl_ids = rannum().choice(
            # low=0,
            high=num_all_pcr_ampl_reads,
            num=num_reads_for_sequencing,
            use_seed=pcr_dict['use_seed'],
            seed=pcr_dict['seed'],
            replace=replace,
        )
        # print(spl_ids)
        df_err_tab = pd.DataFrame(pcr_dict['data'][:, 3:7]).astype(bool)
        df_err_tab.columns = ['mut', 'del', 'ins', 'pcr_err_mark']
        self.console.print("======>+++++++++++bead synthsis error profile++++++++++++")
        self.console.print("======>Before sampling")
        # self.console.print("=========>\n{}".format(df_err_tab))
        # Record it if at least one of them is deletion (True) regardless of whether it comes from bead errors or pcr errors.
        self.console.print("=========>Percentage of reads with deletion errors from bead synthesis: {er}".format(
            er=df_err_tab[df_err_tab['del']].shape[0] / df_err_tab.shape[0],
        ))
        err_union_bead_and_pcr = (df_err_tab['del']) | (df_err_tab['pcr_err_mark'])
        self.console.print("=========>Percentage of reads with deletion errors from bead synthesis and errors from PCR amplication: {er}".format(
            er=err_union_bead_and_pcr[err_union_bead_and_pcr].shape[0]/err_union_bead_and_pcr.shape[0],
        ))
        # print(pcr_dict['data'].shape)
        # print(pcr_dict['data'][:, [1, 2]])
        spl_id_map = tactic6(pcr_dict['data'][:, [1, 2]])
        # print(spl_id_map)
        # print(len(spl_id_map))
        spl_mut_info = pcr_dict['mut_info'][spl_ids]
        # print(pcr_dict['mut_info'].shape)
        df_err_tab_spl = pd.DataFrame(spl_mut_info)[[3, 4, 5, 6]].astype(bool)
        df_err_tab_spl.columns = ['mut', 'del', 'ins', 'pcr_err_mark']
        self.console.print("======>After sampling")
        # self.console.print("=========>\n{}".format(df_err_tab_spl))
        err_union_bead_and_pcr_spl = (df_err_tab_spl['del']) | (df_err_tab_spl['pcr_err_mark'])
        self.console.print("=========>Percentage of reads with deletion errors from bead synthesis: {er}".format(
            er=err_union_bead_and_pcr_spl[err_union_bead_and_pcr_spl].shape[0] / err_union_bead_and_pcr_spl.shape[0],
        ))
        self.console.print("=========>Percentage of reads with deletion errors from bead synthesis and errors from PCR amplication: {er}".format(
            er=df_err_tab_spl[df_err_tab_spl['del']].shape[0] / df_err_tab_spl.shape[0],
        ))
        self.console.print("======>+++++++++++bead synthsis error profile++++++++++++")

        # print(pd.DataFrame(spl_mut_info)[[3, 4, 5]].dtypes)
        # print(len(spl_mut_info))
        keys = spl_mut_info[:, 2]
        # print(len(keys))
        pos_dict = tactic6(pcr_dict['mut_info'][:, [2, 0]])
        base_dict = tactic6(pcr_dict['mut_info'][:, [2, 1]])
        # print(pos_dict)
        # print(base_dict)
        # print(tactic6(base_np))
        res_data = []
        for key in keys:
            mol_id = key.split('_')[0]
            k = key.split('_')[1:]
            # print('kkk', key, k)
            read = umi_map[mol_id]
            # print(read)
            for i in range(len(k)):
                # print('id', i)
                sub_k = mol_id + '_' + '_'.join(k[: i+1]) if k != [] else mol_id
                # print(sub_k)
                # print(pos_dict[sub_k], base_dict[sub_k])
                read = self.errlib.change(read, pos_list=pos_dict[sub_k], base_list=base_dict[sub_k])
            #     print(read)
            # print(read)
            res_data.append([
                read,  # read
                str(mol_id) + '_' + '_'.join(k) if k != [] else str(mol_id),  # sam id
                spl_id_map[str(mol_id) + '_' + '_'.join(k)] if k != [] else 'init',  # source
            ])
            # print(read)
        # print(np.array(res_data).shape)
        return np.array(res_data)

    def sptree(
            self,
            pcr_dict,
            replace=False,
    ):
        """
        short path tree method
        Parameters
        ----------
        pcr_dict

        Returns
        -------

        """
        self.console.print('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
        self.console.print('======>Substitutions of nucleotides by PCR errors using the PCR tree')
        self.console.print('======>Read PCR amplified reads')
        df_seq_lib = pfreader().generic(pcr_dict['read_lib_fpn'])
        # umi_map = df_seq_lib[0].to_dict()
        import pandas as pd
        umi_map = pd.Series(df_seq_lib[0].values, index=df_seq_lib[1].astype(str)).to_dict()
        # @@ single locus
        # {0: 'GGGAAATTTAAACCCTTTAAAGGGAAAAAAGGGCCC', 1: 'GGGTTTAAACCCCCCCCCGGGAAATTTTTTGGGTTT',
        # 2: 'AAATTTGGGCCCGGGAAAGGGCCCAAAAAAGGGAAA', ...,
        # 48: 'CCCAAAAAAGGGCCCAAAGGGCCCTTTGGGGGGCCC', 49: 'TTTTTTAAATTTAAAAAAGGGAAAGGGGGGGGGCCC'}
        # @@ single cell
        # {'0*c*0*g*1*': 'GAAATCATGTAGTTCGGGGGGGCCCTTTTTTTTTTTTAAAAAACCCAAAGGG',
        # '0*c*0*g*4*': 'GAAATCATGTAGTTCGCCCCCCAAATTTAAACCCAAATTTCCCAAAAAAAAA',
        # '1*c*0*g*4*': 'GAAATCATGTAGTTCGAAACCCGGGGGGCCCTTTCCCTTTGGGAAATTTCCC',
        # ...
        # '221*c*1*g*4*': 'CGCGTTAGTAATTCATAAAGGGGGGCCCAAACCCGGGGGGGGGGGGTTTCCC',
        # '222*c*1*g*4*': 'CGCGTTAGTAATTCATAAATTTCCCGGGCCCCCCGGGAAAGGGCCCCCCTTT'}
        # @@ gene
        # {'0*s*0*g*3*': 'GGGGGGCCCTTTTTTTTTTTTAAAAAACCCAAAGGG',
        # '1*s*0*g*3*': 'GGGTTTGGGAAAAAAAAAAAATTTTTTGGGCCCGGG',
        # '2*s*0*g*3*': 'CCCTTTCCCGGGGGGTTTGGGCCCCCCTTTAAAAAA', ...,
        # '421*s*0*g*5*': 'TTTGGGGGGGGGGGGGGGGGGAAAAAAAAAAAATTT',
        # '422*s*0*g*5*': 'AAATTTTTTCCCCCCCCCGGGAAAAAAGGGAAATTT'}

        self.console.print('======>Sampling reads to be sequenced')
        num_all_pcr_ampl_reads = pcr_dict['data'].shape[0]
        self.console.print('=========>There are a total number of {} PCR amplified reads'.format(num_all_pcr_ampl_reads))
        # print(num_all_pcr_ampl_reads)
        # print(pcr_dict.keys())
        if pcr_dict['seq_sub_spl_number'] is not None:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_number']
        else:
            num_reads_for_sequencing = int(pcr_dict['seq_sub_spl_rate'] * num_all_pcr_ampl_reads)
        spl_ids = rannum().choice(
            # low=0,
            high=num_all_pcr_ampl_reads,
            num=num_reads_for_sequencing,
            use_seed=pcr_dict['use_seed'],
            seed=pcr_dict['seed'],
            replace=replace,
        )
        ### spl_ids
        #@@ single locus
        # [ 5221 12317 12284 ...  3697  7994  7549]
        #@@ single cell
        # [  5587  59571  16821 ... 5900  75148]
        # @@ gene
        # [170263  25800 275989 ... 134570 133424]
        ### len(spl_ids)
        # num_reads_for_sequencing
        spl_data = pcr_dict['data'][spl_ids]
        self.console.print('=========>{} reads selected for sequencing'.format(num_reads_for_sequencing))
        ### spl_data
        #@@ single locus
        # [['23_1_2_5_6_8' 'pcr-8']
        #  ['6_3_4_5_8' 'pcr-8']
        #  ['13_1_2_5_7' 'pcr-7']
        #  ...
        #  ['13_3_4_5_7_9' 'pcr-9']
        #  ['4_3_7_9' 'pcr-9']
        #  ['47_2_5_6_7_9' 'pcr-9']]
        #@@ single cell
        # [['77*c*1*g*4*_2_4_5_6_7' 'pcr-7']
        #  ['32*c*1*g*4*_1_2_4_7' 'pcr-7']
        #  ['196*c*1*g*4*_1' 'pcr-1']
        # ...
        #  ['159*c*1*g*4*_1_2_4_5_9' 'pcr-9']
        #  ['49*c*1*g*4*_2_3_4_5_6_7' 'pcr-7']]
        # @@ gene
        # [['261*s*0*g*4*_1_3_5_7_8_9' 'pcr-9']
        #  ['175*s*0*g*4*_1_4_7' 'pcr-7']
        #  ['193*s*0*g*5*_1_2_6_7_10' 'pcr-10']
        # ...
        #  ['182*s*0*g*5*_1_4_8_9' 'pcr-9']
        #  ['226*s*0*g*4*_3_5_6_9' 'pcr-9']]
        ### len(spl_data)
        # num_reads_for_sequencing
        spl_id_map = tactic6(spl_data)
        ### spl_id_map
        #@@ single locus
        # {'23_1_2_5_6_8': 'pcr-8', '6_3_4_5_8': 'pcr-8', '13_1_2_5_7': 'pcr-7', ..., '34_1_2_4': 'pcr-4', '43_1_8': 'pcr-8'}
        #@@ single cell
        # {'77*c*1*g*4*_2_4_5_6_7': 'pcr-7', '32*c*1*g*4*_1_2_4_7': 'pcr-7', '196*c*1*g*4*_1': 'pcr-1', ..., '159*c*1*g*4*_1_2_4_5_9': 'pcr-9', '49*c*1*g*4*_2_3_4_5_6_7': 'pcr-7'}
        #@@ gene
        # {'261*s*0*g*4*_1_3_5_7_8_9': 'pcr-9', '175*s*0*g*4*_1_4_7': 'pcr-7', '193*s*0*g*5*_1_2_6_7_10': 'pcr-10', ..., '182*s*0*g*5*_1_4_8_9': 'pcr-9', '226*s*0*g*4*_3_5_6_9': 'pcr-9'}
        ### len(spl_id_map)
        # num_reads_for_sequencing
        trees = spl_data[:, 0].ravel().tolist()
        ### trees
        #@@ single locus
        # ['23_1_2_5_6_8', '6_3_4_5_8', '13_1_2_5_7', ..., '34_1_2_4', '43_1_8']
        # @@ single cell
        # ['77*c*1*g*4*_2_4_5_6_7', '32*c*1*g*4*_1_2_4_7', '196*c*1*g*4*_1', ..., '159*c*1*g*4*_1_2_4_5_9', '49*c*1*g*4*_2_3_4_5_6_7']
        # @@ gene
        # ['261*s*0*g*4*_1_3_5_7_8_9', '175*s*0*g*4*_1_4_7', '193*s*0*g*5*_1_2_6_7_10', ..., '182*s*0*g*5*_1_4_8_9', '226*s*0*g*4*_3_5_6_9']
        ### len(trees)
        # num_reads_for_sequencing
        self.console.print('=========>Sampled reads contain {} unrepeated PCR amplified molecules'.format(np.unique(trees).size))

        mol_map_to_all_its_pcr_trees = {}
        for tree in trees:
            mol_map_to_all_its_pcr_trees[tree.split('_')[0]] = []
            # print(tree, tree.split('_')[0])
        for tree in trees:
            mol_map_to_all_its_pcr_trees[tree.split('_')[0]].append(tree.split('_')[1:])
        ### mol_map_to_all_its_pcr_trees
        #@@ single locus
        # {'23': [['1', '2', '5', '6', '8'], ['1', '3', '4', '8', '9', '10'],
        # ['1', '4', '7', '9'], ['2', '3', '4', '5', '7', '9'], ['1', '3', '6'],
        # ['1', '2', '4', '7', '8']], '6': [['3', '4', '5', '8'], ['2', '3', '4', '5', '6', '10'],
        # ['5', '6', '7'], ['1', '2', '4', '6', '7', '9'], ['1', '2', '9']],
        # '13': [['1', '2', '5', '7'],
        # ['1', '4', '7', '8', '10'], ['1', '4', '8', '9'], ['6', '7', '9']],
        # ...,
        # '29': [['2', '5', '9', '10'], ['1', '2', '3', '5', '6', '8', '10'],
        # ['3', '4', '7', '9', '10']],
        # '39': [['6', '7', '9']]}
        #@@ single cell
        # {'77*c*1*g*4*': [['2', '4', '5', '6', '7'], ['1', '3', '4', '7', '8', '10'],
        # ['3', '5', '9', '10'], ['1', '2', '3', '5', '6', '7', '10']], '32*c*1*g*4*': [['1', '2', '4', '7'],
        # ['7', '8', '9', '10']], '196*c*1*g*4*': [['1']], ..., '55*c*1*g*4*': [['2', '3', '8', '9']],
        # '49*c*1*g*4*': [['2', '3', '4', '5', '6', '7']]}
        #@@ gene
        # {'261*s*0*g*4*': [['1', '3', '5', '7', '8', '9']], '175*s*0*g*4*':
        # [['1', '4', '7'], ['1', '4', '5', '8', '9']], '193*s*0*g*5*':
        # [['1', '2', '6', '7', '10']], ..., '171*s*0*g*4*': [['2', '3', '4', '6', '8', '9']],
        # '226*s*0*g*4*': [['3', '5', '6', '9']]}
        ### len(mol_map_to_all_its_pcr_trees)
        # number of unique molecules

        mol_sub_tree_map = {}
        for k, v in mol_map_to_all_its_pcr_trees.items():
            mol_sub_tree_map[k] = []
            for j in v:
                mol_sub_tree_map[k].append('_'.join(j))
        ### mol_sub_tree_map
        #@@ single locus
        # {'23': ['1_2_5_6_8', '1_3_4_8_9_10', '1_4_7_9', '2_3_4_5_7_9',
        # '1_3_6', '1_2_4_7_8'], '6': ['3_4_5_8', '2_3_4_5_6_10', '5_6_7',
        # '1_2_4_6_7_9', '1_2_9'], '13': ['1_2_5_7', '1_4_7_8_10', '1_4_8_9',
        # '6_7_9'], ..., '29': ['2_5_9_10', '1_2_3_5_6_8_10', '3_4_7_9_10'], '39': ['6_7_9']}
        #@@ single cell
        # {'77*c*1*g*4*': ['2_4_5_6_7', '1_3_4_7_8_10', '3_5_9_10', '1_2_3_5_6_7_10'],
        #  '32*c*1*g*4*': ['1_2_4_7', '7_8_9_10'], '196*c*1*g*4*': ['1'], ...,
        #  '55*c*1*g*4*': ['2_3_8_9'], '49*c*1*g*4*': ['2_3_4_5_6_7']}
        #@@ gene
        # {'261*s*0*g*4*': ['1_3_5_7_8_9'], '175*s*0*g*4*': ['1_4_7', '1_4_5_8_9'],
        # '193*s*0*g*5*': ['1_2_6_7_10'], ..., '171*s*0*g*4*': ['2_3_4_6_8_9'],
        # '226*s*0*g*4*': ['3_5_6_9']}

        res_data = []
        for k, sub_trees in mol_sub_tree_map.items():
            ### k, sub_trees
            #@@ single locus
            # 23 ['1_2_5_6_8', '1_3_4_8_9_10', '1_4_7_9', '2_3_4_5_7_9', '1_3_6', '1_2_4_7_8']
            # 6 ['3_4_5_8', '2_3_4_5_6_10', '5_6_7', '1_2_4_6_7_9', '1_2_9']
            # 13 ['1_2_5_7', '1_4_7_8_10', '1_4_8_9', '6_7_9']
            # ...
            # 29 ['2_5_9_10', '1_2_3_5_6_8_10', '3_4_7_9_10']
            # 39 ['6_7_9']
            #@@ single cell
            # 77*c*1*g*4* ['2_4_5_6_7', '1_3_4_7_8_10', '3_5_9_10', '1_2_3_5_6_7_10']
            # 32*c*1*g*4* ['1_2_4_7', '7_8_9_10']
            # 196*c*1*g*4* ['1']
            # ...
            # 55*c*1*g*4* ['2_3_8_9']
            # 49*c*1*g*4* ['2_3_4_5_6_7']
            #@@ gene
            # 261*s*0*g*4* ['1_3_5_7_8_9']
            # 175*s*0*g*4* ['1_4_7', '1_4_5_8_9']
            # 193*s*0*g*5* ['1_2_6_7_10']
            # ...
            # 171*s*0*g*4* ['2_3_4_6_8_9']
            # 226*s*0*g*4* ['3_5_6_9']

            read = umi_map[k]
            # read = umi_map[int(k)]
            read_cache = {}
            for sub_tree in sub_trees:
                ### k, sub_tree
                #@@ single locus
                # 23 1_2_5_6_8
                # 23 1_3_4_8_9_10
                # 23 1_4_7_9
                # 23 2_3_4_5_7_9
                # 23 1_3_6
                # 23 1_2_4_7_8
                # 6 3_4_5_8
                # 6 2_3_4_5_6_10
                # 6 5_6_7
                # 6 1_2_4_6_7_9
                # 6 1_2_9
                # 13 1_2_5_7
                # 13 1_4_7_8_10
                # 13 1_4_8_9
                # 13 6_7_9
                # ...
                # 29 2_5_9_10
                # 29 1_2_3_5_6_8_10
                # 29 3_4_7_9_10
                # 39 6_7_9

                #@@ single cell
                # 77*c*1*g*4* ['2_4_5_6_7', '1_3_4_7_8_10', '3_5_9_10', '1_2_3_5_6_7_10']
                # 77*c*1*g*4* 2_4_5_6_7
                # 77*c*1*g*4* 1_3_4_7_8_10
                # ...
                # 49*c*1*g*4* ['2_3_4_5_6_7']
                # 49*c*1*g*4* 2_3_4_5_6_7

                #@@ gene
                # 261*s*0*g*4* 1_3_5_7_8_9
                # 175*s*0*g*4* 1_4_7
                # 175*s*0*g*4* 1_4_5_8_9
                # ...
                # 171*s*0*g*4* 2_3_4_6_8_9
                # 226*s*0*g*4* 3_5_6_9

                if len(sub_tree) > 0:
                    read_id = k + '_' + sub_tree
                else: # sub_tree can be ''. len('') == 0
                    read_id = k
                ### read_id
                #@@ single locus
                # 23_1_2_5_6_8
                # 23_1_3_4_8_9_10
                # 23_1_4_7_9
                # 23_2_3_4_5_7_9
                # 23_1_3_6
                # 23_1_2_4_7_8
                # 6_3_4_5_8
                # 6_2_3_4_5_6_10
                # 6_5_6_7
                # 6_1_2_4_6_7_9
                # 6_1_2_9
                # 13_1_2_5_7
                # 13_1_4_7_8_10
                # 13_1_4_8_9
                # 13_6_7_9
                # ...
                # 29_2_5_9_10
                # 29_1_2_3_5_6_8_10
                # 29_3_4_7_9_10
                # 39_6_7_9

                #@@ single cell
                # 77*c*1*g*4*_2_4_5_6_7
                # 77*c*1*g*4*_1_3_4_7_8_10
                # 77*c*1*g*4*_3_5_9_10
                # ...
                # 55*c*1*g*4*_2_3_8_9
                # 49*c*1*g*4*_2_3_4_5_6_7

                #@@ gene
                # 261*s*0*g*4*_1_3_5_7_8_9
                # 175*s*0*g*4*_1_4_7
                # 175*s*0*g*4*_1_4_5_8_9
                # ...
                # 171*s*0*g*4*_2_3_4_6_8_9
                # 226*s*0*g*4*_3_5_6_9

                k_ = k
                sub_tree_pcr_arr = sub_tree.split('_')
                # print(sub_tree_pcr_arr)
                for pcr in sub_tree_pcr_arr:
                    if pcr != '':
                        k_ = k_ + '_' + pcr
                    ### k, '<======>', k_, 'read id:', read_id
                    ### k, k_
                    #@@ single locus
                    # 23 <======> 23_1 read id: 23_1_2_5_6_8
                    # 23 <======> 23_1_2 read id: 23_1_2_5_6_8
                    # 23 <======> 23_1_2_5 read id: 23_1_2_5_6_8
                    # 23 <======> 23_1_2_5_6 read id: 23_1_2_5_6_8
                    # 23 <======> 23_1_2_5_6_8 read id: 23_1_2_5_6_8
                    # 23 <======> 23_1 read id: 23_1_3_4_8_9_10
                    # 23 <======> 23_1_3 read id: 23_1_3_4_8_9_10
                    # 23 <======> 23_1_3_4 read id: 23_1_3_4_8_9_10
                    # 23 <======> 23_1_3_4_8 read id: 23_1_3_4_8_9_10
                    # 23 <======> 23_1_3_4_8_9 read id: 23_1_3_4_8_9_10
                    # 23 <======> 23_1_3_4_8_9_10 read id: 23_1_3_4_8_9_10
                    # 23 <======> 23_1 read id: 23_1_4_7_9
                    # 23 <======> 23_1_4 read id: 23_1_4_7_9
                    # 23 <======> 23_1_4_7 read id: 23_1_4_7_9
                    # 23 <======> 23_1_4_7_9 read id: 23_1_4_7_9
                    # 23 <======> 23_2 read id: 23_2_3_4_5_7_9
                    # ...

                    #@@ single cell
                    # 55*c*1*g*4* <======> 55*c*1*g*4*_2 read id: 55*c*1*g*4*_2_3_8_9
                    # 55*c*1*g*4* <======> 55*c*1*g*4*_2_3 read id: 55*c*1*g*4*_2_3_8_9
                    # 55*c*1*g*4* <======> 55*c*1*g*4*_2_3_8 read id: 55*c*1*g*4*_2_3_8_9
                    # 55*c*1*g*4* <======> 55*c*1*g*4*_2_3_8_9 read id: 55*c*1*g*4*_2_3_8_9
                    # 49*c*1*g*4* <======> 49*c*1*g*4*_2 read id: 49*c*1*g*4*_2_3_4_5_6_7
                    # 49*c*1*g*4* <======> 49*c*1*g*4*_2_3 read id: 49*c*1*g*4*_2_3_4_5_6_7
                    # 49*c*1*g*4* <======> 49*c*1*g*4*_2_3_4 read id: 49*c*1*g*4*_2_3_4_5_6_7
                    # 49*c*1*g*4* <======> 49*c*1*g*4*_2_3_4_5 read id: 49*c*1*g*4*_2_3_4_5_6_7
                    # 49*c*1*g*4* <======> 49*c*1*g*4*_2_3_4_5_6 read id: 49*c*1*g*4*_2_3_4_5_6_7
                    # 49*c*1*g*4* <======> 49*c*1*g*4*_2_3_4_5_6_7 read id: 49*c*1*g*4*_2_3_4_5_6_7
                    # ...

                    #@@ gene
                    # 261*s*0*g*4* <======> 261*s*0*g*4*_1 read id: 261*s*0*g*4*_1_3_5_7_8_9
                    # 261*s*0*g*4* <======> 261*s*0*g*4*_1_3 read id: 261*s*0*g*4*_1_3_5_7_8_9
                    # 261*s*0*g*4* <======> 261*s*0*g*4*_1_3_5 read id: 261*s*0*g*4*_1_3_5_7_8_9
                    # ...

                    if k_ in read_cache.keys():
                        read = read_cache[k_]
                    else:
                        # print(read)
                        read = self.errlib.mutated(
                            read=read,
                            pcr_error=pcr_dict['pcr_error'],
                        )
                        # print(read)
                    if pcr_dict['pcr_deletion']:
                        read = self.errlib.deletion(read=read, del_rate=pcr_dict['pcr_del_rate'])
                    if pcr_dict['pcr_insertion']:
                        read = self.errlib.insertion(read=read, ins_rate=pcr_dict['pcr_ins_rate'])
                    read_cache[k_] = read
                # print(read_cache)
                res_data.append([
                    read,
                    read_id,
                    spl_id_map[read_id]
                ])
        # print(res_data)
        # print(len(res_data))
        return np.array(res_data)

    def bftree(
            self,
            pcr_dict,
            replace=False,
    ):
        """
        boolean flag tree method

        Notes
        -----
            bool_flags_ = [[] for _ in range(len(uniq_mol_map_new))]
            realvalued_flags_ = [[] for _ in range(len(uniq_mol_map_new))]
            uniq_mol_map_new_ = [[] for _ in range(len(uniq_mol_map_new))]
            for ii, u in enumerate(uniq_mol_map_new):
                for jj, v in enumerate(u):
                    if v != '0':
                        bool_flags_[ii].append(bool_flag_table[ii][jj])
                        realvalued_flags_[ii].append(realvalued_flag_table[ii][jj])
                        uniq_mol_map_new_[ii].append(uniq_mol_map_new[ii][jj])
            print(bool_flags_)
            print(realvalued_flags_)
            print(uniq_mol_map_new_)

        Other Parameters
        ----------------
        trees (i.e., PCR tree)
            [['2', '5', '6', '12', '13', '15'],
             ['3', '4', '7', '9', '12'],
             ['2', '6', '9', '10', '11', '12', '13', '15'],
             ['2', '4', '6', '8', '11'],
             ['3', '4', '7', '9', '13', '14']]
        trees_np
            [['2' '5' '6' '12' '13' '15' '0' '0']
             ['3' '4' '7' '9' '12' '0' '0' '0']
             ['2' '6' '9' '10' '11' '12' '13' '15']
             ['2' '4' '6' '8' '11' '0' '0' '0']
             ['3' '4' '7' '9' '13' '14' '0' '0']]
        bool_flag_table
            [[ True False False False False False False False]
             [ True  True  True  True False False False False]
             [ True False False False False False False False]
             [ True  True False False False False False False]
             [ True  True  True  True False False False False]]
        realvalued_flag_table
            [['2' '-1' '-1' '-1' '-1' '-1' '-1' '-1']
             ['3' '4' '7' '9' '-1' '-1' '-1' '-1']
             ['2' '-1' '-1' '-1' '-1' '-1' '-1' '-1']
             ['2' '4' '-1' '-1' '-1' '-1' '-1' '-1']
             ['3' '4' '7' '9' '-1' '-1' '-1' '-1']]

        Returns
        -------

        """
        self.console.print('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
        self.console.print('======>Substitutions of nucleotides by PCR errors using the PCR tree')
        self.console.print('======>Read PCR amplified reads')
        df_seq_lib = pfreader().generic(pcr_dict['read_lib_fpn'])
        # umi_map = df_seq_lib[0].to_dict()
        import pandas as pd
        umi_map = pd.Series(df_seq_lib[0].values, index=df_seq_lib[1].astype(str)).to_dict()
        # print(umi_map)

        self.console.print('======>Sampling reads to be sequenced')
        num_all_pcr_ampl_reads = pcr_dict['data'].shape[0]
        self.console.print('=========>There are a total number of {} PCR amplified reads'.format(num_all_pcr_ampl_reads))
        # print(num_all_pcr_ampl_reads)
        # print(pcr_dict.keys())
        if pcr_dict['seq_sub_spl_number'] is not None:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_number']
        else:
            num_reads_for_sequencing = pcr_dict['seq_sub_spl_rate'] * num_all_pcr_ampl_reads
        spl_ids = rannum().choice(
            # low=0,
            high=num_all_pcr_ampl_reads,
            num=num_reads_for_sequencing,
            use_seed=pcr_dict['use_seed'],
            seed=pcr_dict['seed'],
            replace=replace,
        )
        ### spl_ids
        # [ 5221 12317 12284 ...  3697  7994  7549]
        ### len(spl_ids)
        # num_reads_for_sequencing

        spl_data = pcr_dict['data'][spl_ids]
        self.console.print('=========>{} reads selected for sequencing'.format(num_reads_for_sequencing))
        ### spl_data
        # [['23_1_2_5_6_8' 'pcr-8']
        #  ['6_3_4_5_8' 'pcr-8']
        #  ['13_1_2_5_7' 'pcr-7']
        #  ...
        #  ['13_3_4_5_7_9' 'pcr-9']
        #  ['4_3_7_9' 'pcr-9']
        #  ['47_2_5_6_7_9' 'pcr-9']]
        ### len(spl_data)
        # num_reads_for_sequencing
        spl_id_map = tactic6(spl_data)
        ### spl_id_map
        # {'23_1_2_5_6_8': 'pcr-8', '6_3_4_5_8': 'pcr-8', '13_1_2_5_7': 'pcr-7', ..., '34_1_2_4': 'pcr-4', '43_1_8': 'pcr-8'}
        ### len(spl_id_map)
        # num_reads_for_sequencing
        trees = spl_data[:, 0].ravel().tolist()
        # print(trees)
        ### trees
        # ['23_1_2_5_6_8', '6_3_4_5_8', '13_1_2_5_7', ... '34_1_2_4', '43_1_8']
        ### len(trees)
        # num_reads_for_sequencing
        self.console.print('=========>Sampled reads contain {} unrepeated PCR amplified molecules'.format(np.unique(trees).size))

        res_data = []
        mol_map_to_all_its_pcr_trees = {}
        for tree in trees:
            mol_map_to_all_its_pcr_trees[tree.split('_')[0]] = []
            # print(tree, tree.split('_')[0])
        for tree in trees:
            mol_map_to_all_its_pcr_trees[tree.split('_')[0]].append(tree.split('_')[1:])
        ### mol_map_to_all_its_pcr_trees
        # {'23': [['1', '2', '5', '6', '8'], ['1', '3', '4', '8', '9', '10'],
        # ['1', '4', '7', '9'], ['2', '3', '4', '5', '7', '9'], ['1', '3', '6'],
        # ['1', '2', '4', '7', '8']], '6': [['3', '4', '5', '8'],
        # ['2', '3', '4', '5', '6', '10'], ['5', '6', '7'], ['1', '2', '4', '6', '7', '9'],
        # ['1', '2', '9']], '13': [['1', '2', '5', '7'], ['1', '4', '7', '8', '10'],
        # ['1', '4', '8', '9'], ['6', '7', '9']],
        # ...
        # '48': [['1'], ['3', '4', '5', '8', '9', '10'], ['1', '9', '10'],
        # ['3', '4', '5', '7', '9', '10']], '11': [['1', '3', '5', '6', '7'],
        # ['2', '3', '4', '7', '8', '10'], ['1', '2', '3', '6', '9']],
        # '29': [['2', '5', '9', '10'], ['1', '2', '3', '5', '6', '8', '10'],
        # ['3', '4', '7', '9', '10']], '39': [['6', '7', '9']]}
        self.console.print('=========>Sampled reads cover {} initial molecules'.format(len(mol_map_to_all_its_pcr_trees)))
        ### len(mol_map_to_all_its_pcr_trees)
        # 50

        for k, trees in mol_map_to_all_its_pcr_trees.items():
            # k, trees
            # 23 [['1', '2', '5', '6', '8'], ['1', '3', '4', '8', '9', '10'], ['1', '4', '7', '9'],
            # ['2', '3', '4', '5', '7', '9'], ['1', '3', '6'], ['1', '2', '4', '7', '8']]
            # 6 [['3', '4', '5', '8'], ['2', '3', '4', '5', '6', '10'], ['5', '6', '7'],
            # ['1', '2', '4', '6', '7', '9'], ['1', '2', '9']]
            # 13 [['1', '2', '5', '7'], ['1', '4', '7', '8', '10'], ['1', '4', '8', '9'], ['6', '7', '9']
            # ...
            # 29 [['2', '5', '9', '10'], ['1', '2', '3', '5', '6', '8', '10'], ['3', '4', '7', '9', '10']]
            # 39 [['6', '7', '9']]
            read = umi_map[str(k)]
            # read
            # CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC
            read_cache_table = [[] for _ in range(len(trees))]
            bool_flag_table = [[True] for _ in range(len(trees))]
            # realvalued_flag_table = [[] for _ in range(len(trees))]
            trees_ori = [[] for _ in range(len(trees))]
            max_len_pcr_tree = max([len(tree) for tree in trees])
            # print(max_len_pcr_tree)
            for _, tree in enumerate(trees):
                if len(tree) < max_len_pcr_tree:
                    tree += ['0' for _ in range(max_len_pcr_tree - len(tree))]
                # k, tree
                # 23 ['1', '2', '5', '6', '8', '0']
                # 23 ['1', '3', '4', '8', '9', '10']
                # 23 ['1', '4', '7', '9', '0', '0']
                # 23 ['2', '3', '4', '5', '7', '9']
                # 23 ['1', '3', '6', '0', '0', '0']
                # 23 ['1', '2', '4', '7', '8', '0']
                # 6 ['3', '4', '5', '8', '0', '0']
                # 6 ['2', '3', '4', '5', '6', '10']
                # 6 ['5', '6', '7', '0', '0', '0']
                # 6 ['1', '2', '4', '6', '7', '9']
                # 6 ['1', '2', '9', '0', '0', '0']
                # 13 ['1', '2', '5', '7', '0']
                # 13 ['1', '4', '7', '8', '10']
                # 13 ['1', '4', '8', '9', '0']
                # 13 ['6', '7', '9', '0', '0']
                # ...
                # 29 ['2', '5', '9', '10', '0', '0', '0']
                # 29 ['1', '2', '3', '5', '6', '8', '10']
                # 29 ['3', '4', '7', '9', '10', '0', '0']
                # 39 ['6', '7', '9']
            trees_np = np.array(trees)
            ### k, trees_np
            # 23 [['1' '2' '5' '6' '8' '0']
            #  ['1' '3' '4' '8' '9' '10']
            #  ['1' '4' '7' '9' '0' '0']
            #  ['2' '3' '4' '5' '7' '9']
            #  ['1' '3' '6' '0' '0' '0']
            #  ['1' '2' '4' '7' '8' '0']]
            # 6 [['3' '4' '5' '8' '0' '0']
            #  ['2' '3' '4' '5' '6' '10']
            #  ['5' '6' '7' '0' '0' '0']
            #  ['1' '2' '4' '6' '7' '9']
            #  ['1' '2' '9' '0' '0' '0']]
            # 13 [['1' '2' '5' '7' '0']
            #  ['1' '4' '7' '8' '10']
            #  ['1' '4' '8' '9' '0']
            #  ['6' '7' '9' '0' '0']]
            # ...
            # 29 [['2' '5' '9' '10' '0' '0' '0']
            #  ['1' '2' '3' '5' '6' '8' '10']
            #  ['3' '4' '7' '9' '10' '0' '0']]
            # 39 [['6' '7' '9']]

            ### +++++++++++++++ block: Construct bool and real-valued flag tables +++++++++++++++
            #         trees (i.e., PCR tree)
            #             [['2', '5', '6', '12', '13', '15'],
            #              ['3', '4', '7', '9', '12'],
            #              ['2', '6', '9', '10', '11', '12', '13', '15'],
            #              ['2', '4', '6', '8', '11'],
            #              ['3', '4', '7', '9', '13', '14']]
            #         trees_np
            #             [['2' '5' '6' '12' '13' '15' '0' '0']
            #              ['3' '4' '7' '9' '12' '0' '0' '0']
            #              ['2' '6' '9' '10' '11' '12' '13' '15']
            #              ['2' '4' '6' '8' '11' '0' '0' '0']
            #              ['3' '4' '7' '9' '13' '14' '0' '0']]
            #         bool_flag_table
            #             [[ True False False False False False False False]
            #              [ True  True  True  True False False False False]
            #              [ True False False False False False False False]
            #              [ True  True False False False False False False]
            #              [ True  True  True  True False False False False]]
            #         realvalued_flag_table
            #             [['2' '-1' '-1' '-1' '-1' '-1' '-1' '-1']
            #              ['3' '4' '7' '9' '-1' '-1' '-1' '-1']
            #              ['2' '-1' '-1' '-1' '-1' '-1' '-1' '-1']
            #              ['2' '4' '-1' '-1' '-1' '-1' '-1' '-1']
            #              ['3' '4' '7' '9' '-1' '-1' '-1' '-1']]
            ### +++++++++++++++ block: construct bool and real-valued flag tables +++++++++++++++
            for id_horiz_in_a_tree in range(trees_np.shape[1]):
                repeat_in_a_col = self.findListDuplicates(trees_np[:, id_horiz_in_a_tree])
                # print('repeated in a col', repeat_in_a_col)
                # print(trees_np[:, id_horiz_in_a_tree])
                for id_vert_across_trees, ele_in_a_col in enumerate(trees_np[:, id_horiz_in_a_tree]):
                    # print('ele_in_a_col {}'.format(ele_in_a_col))
                    if ele_in_a_col in repeat_in_a_col:
                        ids_repeat_in_a_col = [i for i, value in enumerate(trees_np[:, id_horiz_in_a_tree]) if value == ele_in_a_col]
                        # print('asds {}'.format(ids_repeat_in_a_col))
                        if bool_flag_table[id_vert_across_trees][id_horiz_in_a_tree] is False:
                            # print('repeated ele: {}'.format(ele_in_a_col))
                            bool_flag_table[id_vert_across_trees].append(False)
                            # realvalued_flag_table[id_vert_across_trees].append(-1)
                        else:
                            # @@ any positions that is overlay
                            # print('non-repeated ele: {}'.format(ele_in_a_col))
                            inspector_flags = [1 if bool_flag_table[i][id_horiz_in_a_tree] is True else 0 for i in ids_repeat_in_a_col]
                            # print('inspector_flags {}'.format(inspector_flags))
                            if sum(inspector_flags) > 1:
                                bool_flag_table[id_vert_across_trees].append(True)
                                # realvalued_flag_table[id_vert_across_trees].append(ele_in_a_col)
                            else:
                                bool_flag_table[id_vert_across_trees].append(False)
                                # realvalued_flag_table[id_vert_across_trees].append(-1)
                    else:
                        bool_flag_table[id_vert_across_trees].append(False)
                        # realvalued_flag_table[id_vert_across_trees].append(-1)

            bool_flag_table = np.array(bool_flag_table)[:, 1:]
            ### trees_np
            ### bool_flag_table
            # 23
            # [['1' '2' '5' '6' '8' '0']
            #  ['1' '3' '4' '8' '9' '10']
            #  ['1' '4' '7' '9' '0' '0']
            #  ['2' '3' '4' '5' '7' '9']
            #  ['1' '3' '6' '0' '0' '0']
            #  ['1' '2' '4' '7' '8' '0']]
            # [[ True  True False False False False]
            #  [ True  True  True False False False]
            #  [ True False False False False False]
            #  [False False False False False False]
            #  [ True  True False False False False]
            #  [ True  True  True False False False]]
            # 6 [['3' '4' '5' '8' '0' '0']
            #  ['2' '3' '4' '5' '6' '10']
            #  ['5' '6' '7' '0' '0' '0']
            #  ['1' '2' '4' '6' '7' '9']
            #  ['1' '2' '9' '0' '0' '0']]
            # [[False False False False False False]
            #  [False False False False False False]
            #  [False False False False False False]
            #  [ True  True False False False False]
            #  [ True  True False False False False]]
            # 13 [['1' '2' '5' '7' '0']
            #  ['1' '4' '7' '8' '10']
            #  ['1' '4' '8' '9' '0']
            #  ['6' '7' '9' '0' '0']]
            # [[ True False False False False]
            #  [ True  True False False False]
            #  [ True  True False False False]
            #  [False False False False False]]
            # ...
            # 29 [['2' '5' '9' '10' '0' '0' '0']
            #  ['1' '2' '3' '5' '6' '8' '10']
            #  ['3' '4' '7' '9' '10' '0' '0']]
            # [[False False False False False False False]
            #  [False False False False False False False]
            #  [False False False False False False False]]
            # 39 [['6' '7' '9']]
            # [[False False False]]

            for jj in range(trees_np.shape[1]):
                read_for_repeat_tmp_per_col = {}
                for ii, val_in_a_col in enumerate(trees_np[:, jj]):
                    if val_in_a_col != '0':
                        if jj == 0:
                            # print(val_in_a_col, bool_flag_table[ii][jj])
                            if bool_flag_table[ii][jj] == True:
                                if val_in_a_col not in [*read_for_repeat_tmp_per_col.keys()]:
                                    r1 = self.errlib.mutated(
                                        read=read,
                                        pcr_error=pcr_dict['pcr_error'],
                                    )
                                    if pcr_dict['pcr_deletion']:
                                        r1 = self.errlib.deletion(read=r1, del_rate=pcr_dict['pcr_del_rate'])
                                    if pcr_dict['pcr_insertion']:
                                        r1 = self.errlib.insertion(read=r1, ins_rate=pcr_dict['pcr_ins_rate'])
                                    read_for_repeat_tmp_per_col[val_in_a_col] = r1
                                    read_cache_table[ii].append(r1)
                                else:
                                    read_cache_table[ii].append(read_for_repeat_tmp_per_col[val_in_a_col])
                            else:
                                r1 = self.errlib.mutated(
                                    read=read,
                                    pcr_error=pcr_dict['pcr_error'],
                                )
                                if pcr_dict['pcr_deletion']:
                                    r1 = self.errlib.deletion(read=r1, del_rate=pcr_dict['pcr_del_rate'])
                                if pcr_dict['pcr_insertion']:
                                    r1 = self.errlib.insertion(read=r1, ins_rate=pcr_dict['pcr_ins_rate'])
                                read_cache_table[ii].append(r1)
                        if jj > 0:
                            if bool_flag_table[ii][jj] == True:
                                if val_in_a_col + '_' + '_'.join(list(trees_np[ii][:jj])) not in [*read_for_repeat_tmp_per_col.keys()]:
                                    r1 = self.errlib.mutated(
                                        read=read_cache_table[ii][jj - 1],
                                        pcr_error=pcr_dict['pcr_error'],
                                    )[0]
                                    print(r1)
                                    if pcr_dict['pcr_deletion']:
                                        r1 = self.errlib.deletion(read=r1, del_rate=pcr_dict['pcr_del_rate'])
                                    if pcr_dict['pcr_insertion']:
                                        r1 = self.errlib.insertion(read=r1, ins_rate=pcr_dict['pcr_ins_rate'])
                                    read_for_repeat_tmp_per_col[val_in_a_col + '_' + '_'.join(list(trees_np[ii][:jj]))] = r1
                                    read_cache_table[ii].append(r1)
                                else:
                                    read_cache_table[ii].append(read_for_repeat_tmp_per_col[val_in_a_col + '_' + '_'.join(list(trees_np[ii][:jj]))])
                                # print('id', [i for i, value in enumerate(trees_np[:, jj]) if value == val_in_a_col])
                            else:
                                r1 = self.errlib.mutated(
                                    read=read_cache_table[ii][jj - 1],
                                    pcr_error=pcr_dict['pcr_error'],
                                )
                                if pcr_dict['pcr_deletion']:
                                    r1 = self.errlib.deletion(read=r1, del_rate=pcr_dict['pcr_del_rate'])
                                if pcr_dict['pcr_insertion']:
                                    r1 = self.errlib.insertion(read=r1, ins_rate=pcr_dict['pcr_ins_rate'])
                                read_cache_table[ii].append(r1)
            ### k, read_cache_table
            # 23 [['CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC'], ['CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC'], ['CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC'], ['CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC'], ['CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC'], ['CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC', 'CCCCCCGGGGGGAAAAAATTTCCCTTTGGGAAACCC']]
            # 6 [['AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT'], ['AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT'], ['AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT'], ['AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT'], ['AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT', 'AAACCCAAAAAACCCTTTCCCAAACCCGGGGGGTTT']]
            # 13 [['CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG'], ['CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG'], ['CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG'], ['CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG', 'CCCAAATTTTTTAAAGGGTTTGGGCCCGGGGGGGGG']]
            # ...
            # 29 [['TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT'], ['TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT'], ['TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT', 'TTTCCCCCCAAACCCTTTGGGGGGGGGAAAGGGTTT']]
            # 39 [['AAAGGGTTTAAACCCAAAGGGTTTGGGGGGAAAAAA', 'AAAGGGTTTAAACCCAAAGGGTTTGGGGGGAAAAAA', 'AAAGGGTTTAAACCCAAAGGGTTTGGGGGGAAAAAA']]

            for ii, u in enumerate(trees_np):
                for jj, v in enumerate(u):
                    if v != '0':
                        trees_ori[ii].append(trees_np[ii][jj])
            ### k, trees_ori
            # 23 [['1', '2', '5', '6', '8'], ['1', '3', '4', '8', '9', '10'], ['1', '4', '7', '9'], ['2', '3', '4', '5', '7', '9'], ['1', '3', '6'], ['1', '2', '4', '7', '8']]
            # 6 [['3', '4', '5', '8'], ['2', '3', '4', '5', '6', '10'], ['5', '6', '7'], ['1', '2', '4', '6', '7', '9'], ['1', '2', '9']]
            # 13 [['1', '2', '5', '7'], ['1', '4', '7', '8', '10'], ['1', '4', '8', '9'], ['6', '7', '9']]
            # ...
            # 29 [['2', '5', '9', '10'], ['1', '2', '3', '5', '6', '8', '10'], ['3', '4', '7', '9', '10']]
            # 39 [['6', '7', '9']]

            for i, tree in enumerate(trees_ori):
                res_data.append([
                    read_cache_table[i][-1] if read_cache_table[i] != [] else read,  # read
                    str(k) + '_' + '_'.join(tree) if read_cache_table[i] != [] else str(k),  # sample id
                    spl_id_map[str(k) + '_' + '_'.join(tree)] if read_cache_table[i] != [] else 'init',  # source
                ])
            # print('res_data {}'.format(res_data))
        # print(res_data)
        # print(len(res_data))
        return np.array(res_data)

    def findListDuplicates(
            self,
            l
    ):
        seen = set()
        seen_add = seen.add
        seen_twice = set(x for x in l if x in seen or seen_add(x))
        return list(seen_twice)


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd
from functools import wraps
from tresor.util.Console import Console


class Sampling:

    def __init__(
            self,
            method='uniform',
    ):
        self.method = method
        self.console = Console()

    def __call__(self, deal):
        if self.method == 'uniform':
            compiler = self.uniform
        elif self.method == 'gaussian':
            compiler = self.gaussian
        elif self.method == 'poisson':
            compiler = self.poisson
        else:
            compiler = self.uniform
        @wraps(deal)
        def switch(dself, *args, **kwargs):
            # print(args)
            # print(kwargs)
            if 'verbose' in kwargs['params'].keys():
                self.console.verbose = kwargs['params']['verbose']
            else:
                self.console.verbose = True
            self.console.print('======>sampling...')
            res2p = deal(dself, **kwargs)
            # print(res2p)
            res2p['data_spl'] = compiler(
                data=res2p['data'],
                num=res2p['spl_num'],
                # use_seed=res2p['use_seed'],
                # seed=res2p['seed'],
                # replace=res2p['replace'],
            )
            # print(res2p['data'])
            return res2p
        return switch

    def uniform(self, data, num, use_seed=True, seed=1, replace=False):
        """

        Parameters
        ----------
        data
        num
        use_seed
        seed
        replace

        Returns
        -------

        """
        num_samples = len(data)
        if isinstance(data, pd.DataFrame):
            if use_seed:
                state = np.random.RandomState(seed)
                data = data.iloc[state.choice(num_samples, num, replace=replace)].reset_index(drop=True)
            else:
                data = data.iloc[np.random.choice(num_samples, num, replace=replace)].reset_index(drop=True)
        elif type(data) is np.ndarray:
            if use_seed:
                state = np.random.RandomState(seed)
                data = data[state.choice(num_samples, num, replace=replace)]
            else:
                data = data[np.random.choice(num_samples, num, replace=replace)]
        else:
            if use_seed:
                state = np.random.RandomState(seed)
                ids = state.choice(num_samples, num, replace=replace)
                data = [data[i - 1] for i in ids]
            else:
                ids = np.random.choice(num_samples, num, replace=replace)
                data = [data[i - 1] for i in ids]
        return data


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time
import numpy as np
from tresor.library.SingleCell import SingleCell as scsimulib
from tresor.pcr.Amplify import Amplify as pcr
from tresor.pcr.Subsampling import Subsampling
from tresor.sequencing.Calling import Calling as seq
from tresor.util.sequence.fastq.Write import write as wfastq
from tresor.util.file.Folder import Folder as crtfolder
from tresor.util.Console import Console


class SingleCell:

    def __init__(
            self,
            gmat,
            len_params,
            seq_num,
            is_sv_umi_lib,
            is_sv_seq_lib,
            is_sv_primer_lib,
            is_sv_adapter_lib,
            is_sv_spacer_lib,
            working_dir,
            condis,
            sim_thres,
            permutation,

            ampl_rate,
            err_route,
            pcr_error,
            seq_error,
            pcr_num,
            err_num_met,
            use_seed,
            seed,

            sv_fastq_fp,

            seq_sub_spl_number=None,
            seq_sub_spl_rate=1/3,

            verbose=True,
            **kwargs,
    ):
        self.len_params = len_params
        self.umi_unit_lens = self.len_params['umi']['umi_unit_lens']

        self.seq_num = seq_num
        self.is_sv_umi_lib = is_sv_umi_lib
        self.is_sv_seq_lib = is_sv_seq_lib
        self.is_sv_primer_lib = is_sv_primer_lib
        self.is_sv_adapter_lib = is_sv_adapter_lib
        self.is_sv_spacer_lib = is_sv_spacer_lib
        self.working_dir = working_dir
        self.condis = condis
        self.sim_thres = sim_thres
        self.permutation = permutation

        self.gmat = gmat

        self.err_route = err_route
        self.ampl_rate = ampl_rate
        self.pcr_error = pcr_error
        self.pcr_num = pcr_num
        self.err_num_met = err_num_met
        self.use_seed = use_seed
        self.seed = seed

        self.seq_error = seq_error
        self.seq_sub_spl_number = seq_sub_spl_number
        self.seq_sub_spl_rate = seq_sub_spl_rate
        self.sv_fastq_fp = sv_fastq_fp

        self.pcr = pcr
        self.seq = seq
        self.wfastq = wfastq
        self.subsampling = Subsampling()

        self.kwargs = kwargs

        self.console = Console()
        self.console.verbose = verbose
        self.verbose = verbose

    def generate(self, ):
        """

        [['1' '1' '0']
         ['1' '1' '1']
         ['1' '1' '2']
         ...
         ['1' '1' '48']
         ['1' '1' '49']]
        Returns
        -------

        """
        for i, umi_unit_len_i in enumerate(self.umi_unit_lens):
            self.console.print('======>{}. UMI length: {}'.format(i, umi_unit_len_i))
            ### +++++++++++++++ block: generate sequencing library +++++++++++++++
            self.console.print('===>Sequencing library generation starts')
            self.len_params['umi']['umi_unit_len'] = umi_unit_len_i
            working_dir_new = self.working_dir + 'umi_len_' + str(umi_unit_len_i) + '/'
            crtfolder().osmkdir(working_dir_new)
            satime = time.time()
            self.sequencing_library = scsimulib(
                len_params=self.len_params,
                gmat=self.gmat,
                seq_num=self.seq_num,
                is_seed=self.use_seed,
                working_dir=working_dir_new,
                condis=self.condis,
                sim_thres=self.sim_thres,
                permutation=self.permutation,
                is_sv_umi_lib=self.is_sv_umi_lib,
                is_sv_seq_lib=self.is_sv_seq_lib,
                is_sv_primer_lib=self.is_sv_primer_lib,
                is_sv_adapter_lib=self.is_sv_adapter_lib,
                is_sv_spacer_lib=self.is_sv_spacer_lib,
                verbose=self.verbose,
                mode=self.kwargs['mode'],
                material_params=self.kwargs['material_params'] if 'material_params' in self.kwargs.keys() else None,
                seq_params=self.kwargs['seq_params'] if 'seq_params' in self.kwargs.keys() else None,
            ).pooling()
            # print(self.sequencing_library)
            # print(len(self.sequencing_library))
            self.console.print('===>Sequencing library has been generated')

            ### +++++++++++++++ block: PCR amplification: Preparation +++++++++++++++
            self.console.print('===>PCR amplification starts...')
            self.console.print('======>Assign parameters...')
            # print(np.array(self.sequencing_library))
            pcr_ampl_params = {
                'read_lib_fpn': working_dir_new + 'sequencing_library.txt',

                'data': np.array(self.sequencing_library),
                'ampl_rate': self.ampl_rate,
                'pcr_error': self.pcr_error,
                'pcr_num': self.pcr_num,

                'err_route': self.err_route,
                'err_num_met': self.err_num_met,
                'use_seed': self.use_seed,
                'seed': self.seed,
                'recorder_nucleotide_num': [],
                'recorder_pcr_err_num': [],
                'recorder_pcr_read_num': [],

                'seq_sub_spl_number': self.seq_sub_spl_number,
                'seq_sub_spl_rate': self.seq_sub_spl_rate,

                'pcr_deletion': True,
                'pcr_insertion': True, # False True
                'pcr_del_rate': 2.4*10e-6,
                'pcr_ins_rate': 7.1*10e-7,

                'verbose': self.verbose,
            }
            # pcr_ampl_params['data']
            # single-cell
            # [['GAAATCATGTAGTTCGGGGGGGCCCTTTTTTTTTTTTAAAAAACCCAAAGGG' '0*c*0*g*1*' 'init']
            #  ['GAAATCATGTAGTTCGCCCCCCAAATTTAAACCCAAATTTCCCAAAAAAAAA' '0*c*0*g*4*' 'init']
            #  ['GAAATCATGTAGTTCGAAACCCGGGGGGCCCTTTCCCTTTGGGAAATTTCCC' '1*c*0*g*4*' 'init']
            # ...
            # ['CGCGTTAGTAATTCATAAAGGGGGGCCCAAACCCGGGGGGGGGGGGTTTCCC' '221*c*1*g*4*' 'init']
            #  ['CGCGTTAGTAATTCATAAATTTCCCGGGCCCCCCGGGAAAGGGCCCCCCTTT' '222*c*1*g*4*' 'init']]
            if pcr_ampl_params['err_route'] == 'bftree':
                pcr_ampl_params['data'] = pcr_ampl_params['data'][:, 1:3]
            if pcr_ampl_params['err_route'] == 'sptree':
                pcr_ampl_params['data'] = pcr_ampl_params['data'][:, 1:3]
            if pcr_ampl_params['err_route'] == 'mutation_table_minimum' or pcr_ampl_params['err_route'] == 'mutation_table_complete':
                # print(pcr_ampl_params['data'][:, 0])
                def calc_len(a):
                    return len(a)
                vfunc = np.vectorize(calc_len)
                # print(vfunc(pcr_ampl_params['data'][:, 0])[:, np.newaxis])
                # [[52] vfunc(pcr_ampl_params['data'][:, 0])[:, np.newaxis]
                #  [52]
                #  [52]
                #  ...
                #  [52]
                #  [52]]
                pcr_ampl_params['data'] = np.hstack((
                    vfunc(pcr_ampl_params['data'][:, 0])[:, np.newaxis],
                    pcr_ampl_params['data'][:, 1:3],
                ))
                # pcr_ampl_params['data']
                # [['52' '0*c*0*g*1*' 'init']
                #  ['52' '0*c*0*g*4*' 'init']
                #  ['52' '1*c*0*g*4*' 'init']
                #  ...
                #   ['52' '221*c*1*g*4*' 'init']
                #  ['52' '222*c*1*g*4*' 'init']]
                col_0 = np.array([[1] for _ in range(pcr_ampl_params['data'].shape[0])])
                mut_info_table = np.hstack((col_0, col_0))
                col_2 = pcr_ampl_params['data'][:, 1].astype(str)[:, np.newaxis]
                mut_info_table = np.hstack((mut_info_table, col_2))
                pcr_ampl_params['mut_info'] = mut_info_table
                # print(mut_info_table)
                # pcr_ampl_params['mut_info'] = np.empty(shape=[0, 3])
                # print(pcr_ampl_params['mut_info'])

            ### +++++++++++++++ block: PCR amplification: simulation +++++++++++++++
            pcr_stime = time.time()
            pcr = self.pcr(pcr_params=pcr_ampl_params).np()
            # print(pcr.keys())
            self.console.print('======>PCR amplification completes in {}s'.format(time.time() - pcr_stime))

            ### +++++++++++++++ block: Subsampling: sequencing depth or rate +++++++++++++++
            # print(pcr['data'])
            # print(pcr['data'].shape)
            if pcr_ampl_params['err_route'] == 'bftree':
                pcr['data'] = self.subsampling.bftree(pcr_dict=pcr)
            # print(pcr['data'])
            # print(pcr['data'].shape)

            if pcr_ampl_params['err_route'] == 'sptree':
                pcr['data'] = self.subsampling.sptree(pcr_dict=pcr)

            if pcr_ampl_params['err_route'] == 'mutation_table_minimum':
                pcr['data'] = self.subsampling.mutation_table_minimum(pcr_dict=pcr)

            if pcr_ampl_params['err_route'] == 'mutation_table_complete':
                pcr['data'] = self.subsampling.mutation_table_complete(pcr_dict=pcr)

            ### +++++++++++++++ block: Sequencing: parameters +++++++++++++++
            self.console.print('======>Sequencing starts')
            seq_params = {
                'data': pcr['data'],

                'seq_error': self.seq_error,
                'err_num_met': self.err_num_met,
                'use_seed': self.use_seed,
                'seed': self.seed,
                'verbose': self.verbose,

                'seq_sub_spl_number': self.seq_sub_spl_number,
                'seq_sub_spl_rate': self.seq_sub_spl_rate,

                'seq_deletion': True,
                'seq_insertion': True,  # False True
                'seq_del_rate': 2.4 * 10e-6,
                'seq_ins_rate': 7.1 * 10e-7,
            }
            seq = self.seq(seq_params=seq_params).np()
            self.console.print('=========>Sequencing has completed')
            self.console.print('=========>Reads write to files in FastQ format')
            print('======>simulation completes in {}s'.format(time.time() - satime))

            self.wfastq().togz(
                list_2d=seq['data'],
                sv_fp=self.sv_fastq_fp,
                fn='umi_len_' + str(i),
                symbol='-',
            )
            del seq
            self.console.print('=========>FastQ file is saved')
        self.console.print('======>Simulation completes')
        return


if __name__ == "__main__":
    from tresor.path import to

    from tresor.gcell.FromSimulator import fromSimulator

    gmat, _, _ = fromSimulator(
        simulator='spsimseq',
        R_root='D:/Programming/R/R-4.3.2/',
        num_genes=6,
        num_cells=2,
    ).run()

    p = SingleCell(
        # initial sequence generation
        gmat=gmat,

        len_params={
            'umi': {
                'umi_unit_pattern': 3,
                'umi_unit_len': 12,
                'umi_unit_lens': np.arange(7, 36 + 1, 1),
            },
            'umi_1': {
                'umi_unit_pattern': 3,
                'umi_unit_len': 12,
            },
            'barcode': 16,
            'seq': 100,
            'seq_2': 100,
            'adapter': 10,
            'adapter_1': 10,
            'primer': 10,
            'primer_1': 10,
            'spacer': 10,
            'spacer_1': 10,
        },
        seq_params={
            'custom': 'AAGC',
            'custom_1': 'A',
        },
        material_params={
            'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
        },
        seq_num=50,
        working_dir=to('data/simu/'),

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        # condis=['umi'],
        # condis=['umi', 'seq'],
        condis=['barcode', 'umi'],
        # condis=['barcode', 'custom', 'umi', 'custom_1'],
        sim_thres=3,
        permutation=0,

        # PCR amplification
        ampl_rate=0.9,
        err_route='sptree', # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
        pcr_error=1e-4,
        pcr_num=10,
        err_num_met='nbinomial',
        seq_error=0.01,
        seq_sub_spl_number=200, # None
        seq_sub_spl_rate=0.333,
        use_seed=True,
        seed=1,

        verbose=False, # True False

        mode='short_read',  # long_read short_read

        sv_fastq_fp=to('data/simu/'),
    )
    print(p.generate())


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import pandas as pd
from functools import wraps


class Reader(object):

    def __init__(self, ):
        pass

    def __call__(self, deal):
        generic = self.generic
        @wraps(deal)
        def read(ph, *args, **kwargs):
            deal(ph, **kwargs)
            keys = [*kwargs.keys()]
            if kwargs['type'] == 'generic':
                return generic(
                    df_fpn=kwargs['df_fpn'],
                    df_sep='\t' if 'df_sep' not in keys else kwargs['df_sep'],
                    skiprows=False if 'skiprows' not in keys else kwargs['skiprows'],
                    header=None if 'header' not in keys else kwargs['header'],
                    is_utf8=False if 'is_utf8' not in keys else kwargs['is_utf8'],
                )
        return read

    def generic(self, df_fpn, df_sep='\t', skiprows=None, header=None, is_utf8=False):
        if is_utf8:
            return pd.read_csv(
                df_fpn,
                sep=df_sep,
                header=header,
                encoding='utf-8',
                skiprows=skiprows,
            )
        else:
            return pd.read_csv(
                df_fpn,
                sep=df_sep,
                header=header,
                skiprows=skiprows
            )


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from tresor.sequencing.Error import Error as seqerr
from tresor.util.random.Sampling import Sampling as ranspl


class Calling:

    def __init__(self, seq_params):
        self.seq_params = seq_params

    def np(self, ):
        if self.seq_params['seq_sub_spl_number']:
            self.seq_params['spl_num'] = self.seq_params['seq_sub_spl_number']
        else:
            self.seq_params['spl_num'] = int(self.seq_params['data'].shape[0] * self.seq_params['seq_sub_spl_rate'])
        res_flow = self.flow(params=self.seq_params)
        # print(res_flow)
        return res_flow

    @seqerr(method='default')
    @ranspl(method='uniform')
    def flow(self, params):
        return params


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"



def tactic6(arr_2d):
    result = {}
    len_arr = len(arr_2d[0])
    if len_arr == 2:
        for item in arr_2d:
            result[item[0]] = item[1]
    else:
        for item in arr_2d:
            result[item[0]] = item[1:]
    return result



__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import click
from pyfiglet import Figlet

### @@@ gmat_bulk gmat_sc
from tresor.gmat import spsimseq_bulk as gmat_spsimseq_bulk
from tresor.gmat import spsimseq_sc as gmat_spsimseq_sc

### library_sl | library_bulk | library_sc
from tresor.locus import library as lib_sl
from tresor.gene import library as lib_bulk
from tresor.sc import library as lib_sc

### seqerr_sl | pcrerr_sl | pcrnum_sl | amplrate_sl | umilen_sl | seqdep_sl | generic_sl
from tresor.locus import simu_seq_err as seqerr_sl
from tresor.locus import simu_pcr_err as pcrerr_sl
from tresor.locus import simu_pcr_num as pcrnum_sl
from tresor.locus import simu_ampl_rate as amplrate_sl
from tresor.locus import simu_umi_len as umilen_sl
from tresor.locus import simu_seq_dep as seqdep_sl
from tresor.locus import simu_generic as generic_sl

### seqerr_gene | pcrerr_gene | pcrnum_gene | amplrate_gene | umilen_gene | seqdep_gene
from tresor.gene import simu_seq_err as seqerr_gene
from tresor.gene import simu_pcr_err as pcrerr_gene
from tresor.gene import simu_pcr_num as pcrnum_gene
from tresor.gene import simu_ampl_rate as amplrate_gene
from tresor.gene import simu_umi_len as umilen_gene
from tresor.gene import simu_seq_dep as seqdep_gene

### seqerr_sc | pcrerr_sc | pcrnum_sc | amplrate_sc | umilen_sc | seqdep_sc
from tresor.sc import simu_seq_err as seqerr_sc
from tresor.sc import simu_pcr_err as pcrerr_sc
from tresor.sc import simu_pcr_num as pcrnum_sc
from tresor.sc import simu_ampl_rate as amplrate_sc
from tresor.sc import simu_umi_len as umilen_sc
from tresor.sc import simu_seq_dep as seqdep_sc

from tresor.util.Console import Console


vignette1 = Figlet(font='slant')

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

console = Console()
console.verbose = True


class HelpfulCmd(click.Command):

    def format_help(self, ctx, formatter):
        click.echo(vignette1.renderText('Tresor'))
        click.echo(
            """
            tool 
                gmat_bulk | gmat_sc
                
                @@@ gmat_bulk
                tresor gmat_bulk -rfpn D:/Programming/R/R-4.3.2/ -nspl 2 -ngene 10 -gsimulator spsimseq -wd ./tresor/data/spsimseq_bulk.h5 -is True -vb True
            
                @@@ gmat_sc
                tresor gmat_sc -rfpn D:/Programming/R/R-4.3.2/ -ncell 10 -ngene 10 -gsimulator spsimseq -wd ./tresor/data/spsimseq_sc.h5 -is True -vb True
            
            
                library_sl | library_bulk | library_sc
                
                @@@ library_sl
                tresor library_sl -cfpn ./tresor/data/libslocus.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
                
                @@@ library_bulk
                tresor library_bulk -cfpn ./tresor/data/libgene.yml -snum 50 -rfpn D:/Programming/R/R-4.3.2/ -nspl 2 -ngene 20 -gsimulator spsimseq -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True 

                @@@ library_sc
                tresor library_sc -cfpn ./tresor/data/libsc.yml -snum 50 -rfpn D:/Programming/R/R-4.3.2/ -ncell 10 -ngene 10 -gsimulator spsimseq -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True 

                
                seqerr_sl | pcrerr_sl | pcrnum_sl | amplrate_sl | umilen_sl | seqdep_sl | generic_sl
                
                @@@ seqerr_sl
                tresor seqerr_sl -cfpn ./tresor/data/seqerr_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ pcrerr_sl
                tresor pcrerr_sl -cfpn ./tresor/data/pcrerr_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ pcrnum_sl
                tresor pcrnum_sl -cfpn ./tresor/data/pcrnum_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ amplrate_sl
                tresor amplrate_sl -cfpn ./tresor/data/amplrate_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ umilen_sl
                tresor umilen_sl -cfpn ./tresor/data/umilen_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ seqdep_sl
                tresor seqdep_sl -cfpn ./tresor/data/seqdep_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ generic_sl
                tresor generic_sl -cfpn ./tresor/data/generic_sl.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
            
                seqerr_gene | pcrerr_gene | pcrnum_gene | amplrate_gene | umilen_gene | seqdep_gene
                @@@ seqerr_gene
                tresor seqerr_gene -cfpn ./tresor/data/seqerr_gene.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ pcrerr_gene
                tresor pcrerr_gene -cfpn ./tresor/data/pcrerr_gene.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ pcrnum_gene
                tresor pcrnum_gene -cfpn ./tresor/data/pcrnum_gene.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ amplrate_gene
                tresor amplrate_gene -cfpn ./tresor/data/amplrate_gene.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ umilen_gene
                tresor umilen_gene -cfpn ./tresor/data/umilen_gene.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ seqdep_gene
                tresor seqdep_gene -cfpn ./tresor/data/seqdep_gene.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
            
                seqerr_sc | pcrerr_sc | pcrnum_sc | amplrate_sc | umilen_sc | seqdep_sc
                @@@ seqerr_sc
                tresor seqerr_sc -cfpn ./tresor/data/seqerr_sc.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ pcrerr_sc
                tresor pcrerr_sc -cfpn ./tresor/data/pcrerr_sc.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ pcrnum_sc
                tresor pcrnum_sc -cfpn ./tresor/data/pcrnum_sc.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ amplrate_sc
                tresor amplrate_sc -cfpn ./tresor/data/amplrate_sc.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ umilen_sc
                tresor umilen_sc -cfpn ./tresor/data/umilen_sc.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                @@@ seqdep_sc
                tresor seqdep_sc -cfpn ./tresor/data/seqdep_sc.yml -snum 50 -permut 0 -sthres 3 -wd ./tresor/data/simu/ -md short_read -is True -vb True
            
                
            """
        )


@click.command(cls=HelpfulCmd, context_settings=CONTEXT_SETTINGS)
@click.argument('tool', type=str)
@click.option(
    '-cfpn', '--config_fpn', type=str,
    # required=True,
    help="""
        Path to a YMAL file
    """
)
@click.option(
    '-wd', '--working_dir', type=str, required=True,
    help="""
        Path to store results in the working directory
    """
)
@click.option(
    '-snum', '--seq_num', type=int,
    help="""
        read/UMI number
    """
)
@click.option(
    '-sthres', '--sim_thres', type=int,
    help="""
        Similarity threshold between UMIs
    """
)
@click.option(
    '-permut', '--permutation', type=int,
    help="""
        permutation 
    """
)
@click.option(
    '-md', '--mode', type=str,
    help="""
        short_read or long_read
    """
)
@click.option(
    '-is', '--is_seed', type=bool, default=True,
    help="""
        permutation 
    """
)
@click.option(
    '-isv_umi', '--is_sv_umi_lib', type=bool, default=True,
    help="""
        if is it save UMI library 
    """
)
@click.option(
    '-isv_seq', '--is_sv_seq_lib', type=bool, default=True,
    help="""
        if is it save sequence library 
    """
)
@click.option(
    '-isv_primer', '--is_sv_primer_lib', type=bool, default=True,
    help="""
        if is it save primer library 
    """
)
@click.option(
    '-isv_adapter', '--is_sv_adapter_lib', type=bool, default=True,
    help="""
        if is it save adapter library 
    """
)
@click.option(
    '-isv_spacer', '--is_sv_spacer_lib', type=bool, default=True,
    help="""
        if is it save spacer library 
    """
)
@click.option(### @@@ library bulk-RNA-seq simulation params
    '-rfpn', '--r_root', type=str,
    help="""
        R root directory path
    """
)
@click.option(
    '-nspl', '--num_samples', type=int,
    help="""
        number of samples
    """
)
@click.option(
    '-ngene', '--num_genes', type=int,
    help="""
        number of genes
    """
)
@click.option(
    '-gsimulator', '--gmat_simulator', type=str, default="spsimseq",
    help="""
        gmat simulator
    """
)

@click.option(
    '-ncell', '--num_cells', type=int,
    help="""
        number of cells
    """
)

@click.option(
    '-vb', '--verbose', type=bool, default=True,
    help="""
        Print verbose output
    """
)
def main(
        tool,
        config_fpn,
        working_dir,
        seq_num,

        # gspl
        r_root,
        num_samples,
        num_genes,
        gmat_simulator,

        # gmat
        num_cells,

        sim_thres,
        permutation,
        is_seed,
        is_sv_umi_lib,
        is_sv_seq_lib,
        is_sv_primer_lib,
        is_sv_adapter_lib,
        is_sv_spacer_lib,
        mode,
        verbose,
):
    print(vignette1.renderText('Tresor'))
    ### @@@ simu library
    if tool == "library_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        lib_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
        )
    elif tool == "library_bulk":
        console.print("=============>Tool {} is being used...".format(tool))
        lib_bulk(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            r_root=r_root,
            num_samples=num_samples,
            num_genes=num_genes,
            simulator=gmat_simulator,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
        )
    elif tool == "library_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        lib_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            r_root=r_root,
            num_cells=num_cells,
            num_genes=num_genes,
            simulator=gmat_simulator,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
        )

    ### @@@ simu gmat
    elif tool == "gmat_bulk":
        console.print("=============>Tool {} is being used...".format(tool))
        if gmat_simulator == 'spsimseq':
            gmat_spsimseq_bulk(
                R_root=r_root,
                num_samples=num_samples,
                num_genes=num_genes,
                simulator=gmat_simulator,
                sv_fpn=working_dir,
            )
    elif tool == "gmat_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        if gmat_simulator == 'spsimseq':
            gmat_spsimseq_sc(
                R_root=r_root,
                num_cells=num_cells,
                num_genes=num_genes,
                simulator=gmat_simulator,
                sv_fpn=working_dir,
            )
    ### @@@ simu single locus
    elif tool == "seqerr_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        seqerr_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "pcrerr_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        pcrerr_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "pcrnum_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        pcrnum_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "amplrate_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        amplrate_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "umilen_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        umilen_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "seqdep_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        seqdep_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "generic_sl":
        console.print("=============>Tool {} is being used...".format(tool))
        generic_sl(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )

    ### @@@ simu bulk
    elif tool == "seqerr_gene":
        console.print("=============>Tool {} is being used...".format(tool))
        seqerr_gene(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "pcrerr_gene":
        console.print("=============>Tool {} is being used...".format(tool))
        pcrerr_gene(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "pcrnum_gene":
        console.print("=============>Tool {} is being used...".format(tool))
        pcrnum_gene(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "amplrate_gene":
        console.print("=============>Tool {} is being used...".format(tool))
        amplrate_gene(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "umilen_gene":
        console.print("=============>Tool {} is being used...".format(tool))
        umilen_gene(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "seqdep_gene":
        console.print("=============>Tool {} is being used...".format(tool))
        seqdep_gene(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )

    ### @@@ simu sc
    elif tool == "seqerr_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        seqerr_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "pcrerr_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        pcrerr_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "pcrnum_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        pcrnum_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "amplrate_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        amplrate_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "umilen_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        umilen_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )
    elif tool == "seqdep_sc":
        console.print("=============>Tool {} is being used...".format(tool))
        seqdep_sc(
            config_fpn=config_fpn,
            working_dir=working_dir,
            seq_num=seq_num,
            sim_thres=sim_thres,
            permutation=permutation,
            is_seed=is_seed,
            is_sv_umi_lib=is_sv_umi_lib,
            is_sv_seq_lib=is_sv_seq_lib,
            is_sv_primer_lib=is_sv_primer_lib,
            is_sv_adapter_lib=is_sv_adapter_lib,
            is_sv_spacer_lib=is_sv_spacer_lib,
            mode=mode,
            verbose=verbose,
            sv_fastq_fp=working_dir,
        )


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from tresor.util.Console import Console


class fromSimulator:

    def __init__(
            self,
            simulator,
            R_root,
            num_samples=2,
            num_genes=20,
            verbose=True,
    ):
        self.simulator = simulator
        self.num_samples = num_samples
        self.num_genes = num_genes
        self.R_root = R_root

        self.console = Console()
        self.console.verbose = verbose

    def spsimseq(self, ):
        import os
        os.environ['R_HOME'] = self.R_root
        import rpy2.robjects as rob
        from rpy2.robjects import pandas2ri
        from rpy2.robjects.conversion import localconverter
        self.console.print('=========>spsimseq is being used')
        res = rob.r(
            """
            suppressPackageStartupMessages(library(SPsimSeq))
            cat("SPsimSeq package version", as.character(packageVersion("SPsimSeq")), "\n")
            
            # load the Zhang bulk RNA-seq data
            data("zhang.data.sub")
            # filter genes with sufficient expression (important step) 
            zhang.counts <- zhang.data.sub$counts[rowSums(zhang.data.sub$counts > 0)>=5, ]
            
            set.seed(6452)
            zhang.counts2 <- zhang.counts[sample(nrow(zhang.counts), 20), ]
            sim.data.bulk <- SPsimSeq(n.sim = 1, s.data = zhang.counts2,
                                      group = zhang.data.sub$MYCN.status, 
                                      n.genes = {},
            """.format(self.num_genes)+""" 
                                      batch.config = 1,
                                      group.config = c(0.5, 0.5), tot.samples = {},
                                      """.format(self.num_samples)+"""
                                      pDE = 0.5, lfc.thrld = 0.5, 
                                      result.format = "list")
            sim.data.bulk1 <- sim.data.bulk[[1]]                              
            return (data.frame(sim.data.bulk1$counts))
            """
        )
        self.console.print('=========>spsimseq completes simulation')
        with localconverter(rob.default_converter + pandas2ri.converter):
            df = rob.conversion.rpy2py(res)
        return df.T

    def tool(self, ):
        return {
            'spsimseq': self.spsimseq()
        }

    def run(self, ):
        return self.tool()[self.simulator]


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import numpy as np
from functools import wraps
from tresor.util.Console import Console


class Number:

    def __init__(
            self,
            *args,
            **kwargs,
    ):
        self.args = args
        self.kwargs = kwargs
        self.console = Console()

    def __call__(self, deal):
        if self.kwargs['type'] == 'binomial':
            distrib = self.binomial
        if self.kwargs['type'] == 'uniform':
            distrib = self.binomial
        else:
            pass
        @wraps(deal)
        def switch(ph, *args, **kwargs):
            if 'verbose' in kwargs['params'].keys():
                self.console.verbose = kwargs['params']['verbose']
            else:
                self.console.verbose = True
            self.console.print('======>numbering...')
            # print(kwargs)
            res = deal(ph, **kwargs)
            res['spl_num'] = distrib(
                n=len(res['data']),
                p=res['ampl_rate'],
            )
            # print(res)
            return res
        return switch

    def binomial(self, n, p, use_seed=True, seed=1):
        if use_seed:
            state = np.random.RandomState(seed)
            return state.binomial(
                n,
                p,
            )
        else:
            return np.random.binomial(
                n,
                p,
            )

    def nbinomial(self, n, p, use_seed=True, seed=1):
        """

        Parameters
        ----------
        n
            the number of success to be expected, better, n = the total number of trails * p
        p
            the prob of success
        use_seed
        seed

        Returns
        -------

        """
        if use_seed:
            state = np.random.RandomState(seed)
            return state.negative_binomial(
                n,
                p,
            )
        else:
            return np.random.negative_binomial(
                n,
                p,
            )

    def uniform(self, low, high, num, use_seed=True, seed=1):
        if use_seed:
            state = np.random.RandomState(seed)
            return state.randint(
                low=low,
                high=high,
                size=num
            )
        else:
            return np.random.randint(
                low=low,
                high=high,
                size=num
            )

    # def choice(self, high, num, replace=False):
    #     from numpy.random import default_rng
    #     rng = default_rng()
    #     return rng.choice(high, size=num, replace=replace)

    def choice(self, high, num, use_seed=True, seed=1, replace=False):
        if use_seed:
            state = np.random.RandomState(seed)
            return state.choice(high, int(num), replace=replace)
        else:
            return np.random.choice(high, int(num), replace=replace)


__version__ = "v1.0"
__copyright__ = "Copyright 2023"
__license__ = "MIT"
__lab__ = "cribbslab"

import time
import numpy as np
from simreadflow.util.random.Sampling import sampling as ranspl
from simreadflow.util.random.Number import number as rannum
from simreadflow.util.file.read.Reader import reader as pfreader
from simreadflow.util.file.create.Folder import folder as crtfolder
from simreadflow.util.sequence.symbol.Single import single as dnasgl
from simreadflow.read.umi.Design import design as dumi
from simreadflow.read.similarity.distance.Hamming import hamming


class umiDouble(object):

    def __init__(self, seq_num, is_seed=False, umi_unit_pattern=3, umi_unit_len=12, is_sv_umi_lib=True, umi_lib_fpn='./umi.txt', working_dir='./simu/', condis=['umi'], sim_thres=2, permutation=0):
        self.pfreader = pfreader()
        self.ranspl = ranspl()
        self.rannum = rannum()
        self.dnasgl = dnasgl()
        self.crtfolder = crtfolder()
        self.dumi = dumi
        self.is_seed = is_seed
        self.is_sv_umi_lib = is_sv_umi_lib
        self.umi_lib_fpn = umi_lib_fpn
        self.seq_num = seq_num
        self.umi_unit_pattern = umi_unit_pattern
        self.umi_unit_len = umi_unit_len
        self.condis = condis
        self.sim_thres = sim_thres
        self.permutation = permutation
        self.dna_map = self.dnasgl.todict(nucleotides=self.dnasgl.get(universal=True), reverse=True)
        self.crtfolder.osmkdir(working_dir)

    @property
    def umi_len(self, ):
        return self.umi_unit_pattern * self.umi_unit_len

    def pooling(self,):
        stime = time.time()
        seqs = []
        umi_pool = []
        umi_cnt = 0
        for id in np.arange(self.seq_num):
            read_struct_ref = {}
            if 'umi' in self.condis:
                umi_flag = False
                while not umi_flag:
                    umip = self.dumi(
                        dna_map=self.dna_map,
                        umi_unit_pattern=self.umi_unit_pattern,
                        pseudorandom_num=self.rannum.uniform(
                            low=0,
                            high=4,
                            num=self.umi_unit_len,
                            use_seed=self.is_seed,
                            seed=id + self.permutation * self.seq_num + umi_cnt,
                        ),
                    )
                    umi_i = umip.reoccur(is_sv=False)
                    edh = np.array([hamming().general(umi_i, j) for j in umi_pool])
                    # for j in umi_pool:
                    #     if hamming().general(umi_i, j) < self.sim_thres:
                    #         print(umi_i, j)
                    if len(edh[edh < self.sim_thres]) == 0:
                        # print(len(edh[edh < self.sim_thres]))
                        umi_pool.append(umi_i)
                        read_struct_ref['umi'] = umi_i
                        umi_flag = True
                        umip.write(res=umi_i, lib_fpn=self.umi_lib_fpn, is_sv=self.is_sv_umi_lib)
                    else:
                        # print(id)
                        umi_cnt += 1
            read_struct_pfd_order = {condi: read_struct_ref[condi] for condi in self.condis}
            seqs.append([self.paste([*read_struct_pfd_order.values()]), id])
        print(umi_cnt)
        # print(umi_pool)

        etime = time.time()
        print("===>time for generating initial pool of sequences: {:.3f}s".format(etime-stime))
        print(seqs)
        reads = list(np.reshape(seqs, (int(self.seq_num / 2), 4)))
        c = []
        pos_transloc = np.random.choice(int(self.seq_num / 2), int(0.1 * self.seq_num), replace=False)
        print(pos_transloc)
        for i, r in enumerate(reads):
            mark = 'real_yes' if i in pos_transloc else 'real_no'
            # ===> 'read', 'r1_id', 'r2_id', 'transloc_stat', 'transloc_side', 'sam_id', 'source'
            c.append([r[0] + r[2], r[1], r[3], mark, 'none', i, 'init'])
        return c

    def paste(self, read_struct=[]):
        return ''.join(read_struct)


if __name__ == "__main__":
    from Path import to
    DEFINE = {
        '': '',
    }
    # print(DEFINE['cand_pool_fpn'])
    p = umiDouble(
        seq_num=50,
        umi_unit_pattern=3,
        umi_unit_len=12,
        is_seed=True,

        is_sv_umi_lib=False,
        umi_lib_fpn=to('data/simu/transloc/trimer/umi.txt'),
        working_dir=to('data/simu/transloc/trimer/'),

        condis=['umi'],
        sim_thres=3,
        permutation=1,
    )

    # print(p.umi_len)
    res = p.pooling()
    print(res)